

import math, datetime

from decimal import Decimal, ROUND_UP
from collections import namedtuple

from .. import tag

from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class Chart1(Widget):

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'values':FieldArgList('integer', jsonset=True),
                        'stroke_width':FieldArg("integer", 1),
                        'stroke':FieldArg("text", "black"),
                        'plus50legend':FieldArg("text", ""),
                        'minus50legend':FieldArg("text", ""),
                        'zerolegend':FieldArg("text", ""),
                        'font_family':FieldArg("text", "arial")
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a chart, held in a 200 high x 500 wide space
           The vertical scale is +100 to -100
           values: a list of integers, each integer should be between -100 and +100
           with a maximum of fifty integers.
           These will be plotted on the chart at 10point intervals, with the
           last point on the right of the chart
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self[0] = tag.ClosedPart(tag_name='rect', attribs={"x":"0",
                                                           "y":"0",
                                                           "width":"500",
                                                           "height":"200",
                                                           "fill":"white",
                                                           "stroke":"green",
                                                           "stroke-width":"1"})
        # centre line
        self[1] = tag.ClosedPart(tag_name='line', attribs={"x1":"0",
                                                           "y1":"100",
                                                           "x2":"500",
                                                           "y2":"100",
                                                           "stroke":"green",
                                                           "stroke-width":"3"})

        # 50 lines - horizontal
        self[2] = tag.ClosedPart(tag_name='line', attribs={"x1":"0",
                                                           "y1":"50",
                                                           "x2":"500",
                                                           "y2":"50",
                                                           "stroke":"green",
                                                           "stroke-width":"1"})
        self[3] = tag.ClosedPart(tag_name='line', attribs={"x1":"0",
                                                           "y1":"150",
                                                           "x2":"500",
                                                           "y2":"150",
                                                           "stroke":"green",
                                                           "stroke-width":"1"})
        # 50 lines - vertical
        for n in range(50, 500, 50):
            self.append(tag.ClosedPart(tag_name='line', attribs={"x1":str(n),
                                                                 "y1":"0",
                                                                 "x2":str(n),
                                                                 "y2":"200",
                                                                 "stroke":"green",
                                                                 "stroke-width":"1"}))


    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        font_family = self.get_field_value("font_family")
        if not font_family:
            font_family = "arial"

        if self.get_field_value("plus50legend"):
            charnumbers = len(self.get_field_value("plus50legend"))
            textlength = 5*charnumbers
            self.append(tag.ClosedPart(tag_name='rect', attribs={"x":"3",
                                                           "y":"40",
                                                           "width":str(textlength+10),
                                                           "height":"20",
                                                           "fill":"white",
                                                           "stroke-width":"0"}))
            self.append( tag.Part(tag_name='text',
                                  text=self.get_field_value("plus50legend"),
                                  attribs={
                                            'x':"5",
                                            'y':"53",
                                            'font-size': '10',
                                            'font-family': font_family,
                                            'lengthAdjust':"spacingAndGlyphs",
                                            'textLength':str(textlength),
                                            'fill':"green",
                                            'stroke-width':"0"  }))

        if self.get_field_value("minus50legend"):
            charnumbers = len(self.get_field_value("minus50legend"))
            textlength = 5*charnumbers
            self.append(tag.ClosedPart(tag_name='rect', attribs={"x":"3",
                                                           "y":"140",
                                                           "width":str(textlength+10),
                                                           "height":"20",
                                                           "fill":"white",
                                                           "stroke-width":"0"}))
            self.append( tag.Part(tag_name='text',
                                  text=self.get_field_value("minus50legend"),
                                  attribs={
                                            'x':"5",
                                            'y':"153",
                                            'font-size': '10',
                                            'font-family': font_family,
                                            'lengthAdjust':"spacingAndGlyphs",
                                            'textLength':str(textlength),
                                            'fill':"green",
                                            'stroke-width':"0"  }))


        if self.get_field_value("zerolegend"):
            charnumbers = len(self.get_field_value("zerolegend"))
            textlength = 5*charnumbers
            self.append(tag.ClosedPart(tag_name='rect', attribs={"x":"3",
                                                           "y":"90",
                                                           "width":str(textlength+10),
                                                           "height":"20",
                                                           "fill":"white",
                                                           "stroke-width":"0"}))
            self.append( tag.Part(tag_name='text',
                                  text=self.get_field_value("zerolegend"),
                                  attribs={
                                            'x':"5",
                                            'y':"103",
                                            'font-size': '10',
                                            'font-family': font_family,
                                            'lengthAdjust':"spacingAndGlyphs",
                                            'textLength':str(textlength),
                                            'fill':"green",
                                            'stroke-width':"0"  }))

        stroke_width = self.get_field_value("stroke_width")
        if not stroke_width:
            stroke_width = "1"
        else:
            stroke_width = str(stroke_width)

        stroke = self.get_field_value("stroke")
        if not stroke:
            stroke = "black"

        values = self.get_field_value("values")
        if not values:
            # still include a polyline tag, for javascript to find
            self.append(tag.ClosedPart(tag_name='polyline', attribs={"points":"",
                                                                     "fill":"none",
                                                                     "stroke":stroke,
                                                                     "stroke-width":stroke_width}))

            return

        vals = list(reversed(values))

        xpoint = 510
        points = ""
        for ypoint in vals:
            xpoint = xpoint-10
            if xpoint < 0:
                break
            if ypoint > 100:
                y = "0"
            elif ypoint < -100:
                y = "200"
            else:
                y = str(100-ypoint)
            points = points + " " + str(xpoint) + "," + y

        self.append(tag.ClosedPart(tag_name='polyline', attribs={"points":points,
                                                                 "fill":"none",
                                                                 "stroke":stroke,
                                                                 "stroke-width":stroke_width}))

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <rect /> <!-- the chart rectangle -->
  <!-- lines and text which draw the chart -->
  <polyline /> <!-- Draws the values on the chart -->
</g>"""




class Graph48Hr(Widget):

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'values':FieldArgTable(('text', 'datetime')),
                        'fill':FieldArg("text", "white"),
                        'fill_opacity':FieldArg("text", "1"),
                        'plotcol':FieldArg("text", "black"),
                        'font_family':FieldArg("text", "arial"),
                        'axiscol':FieldArg("text", "green"),
                        'minvalue':FieldArg("text", "0"),
                        'maxvalue':FieldArg("text", "100"),
                        'last_day':FieldArg("date", "")
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a graph, held in a 1200 high x 1400 wide space
           values: a list of datetime objects covering a 48 hr period
           These will be plotted on the chart to the latest date in the values
           If last_day given, rightmost axis will be midnight of that day 
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self._hr = datetime.timedelta(hours=1)
        self._axiscol = "green"
        self._plotcol = "black"
        self._font_family = "arial"


    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        fill = self.get_field_value("fill")
        if not fill:
            fill = "white"

        fill_opacity = self.get_field_value("fill_opacity")
        if not fill_opacity:
            fill_opacity = "0"

        self._font_family = self.get_field_value("font_family")
        if not self._font_family:
            self._font_family = "arial"

        self._axiscol = self.get_field_value("axiscol")
        if not self._axiscol:
            self._axiscol = "green"

        self._plotcol = self.get_field_value("plotcol")
        if not self._plotcol:
            self._plotcol = "black"

        self[0] = tag.ClosedPart(tag_name='rect', attribs={"x":"240",
                                                           "y":"20",
                                                           "width":"960",
                                                           "height":"720",
                                                           "fill":fill,
                                                           "fill-opacity":fill_opacity,
                                                           "stroke":self._axiscol,
                                                           "stroke-width":"1"})

        values = self.get_field_value("values")
        last_day = self.get_field_value("last_day")

        if (not values) and (not last_day):
            # do not know when to plot
            return

        # get maxt the last datetime to plot
        if last_day:
            # set maxt at midnight of last_day
            maxt = datetime.datetime(last_day.year, last_day.month, last_day.day) + datetime.timedelta(days=1)
        else:
            # set maxt as the start of the hour after the last plotted value
            maxt = values[0][1]
            for valpair in values:
                val, t = valpair
                # t is a datetime object
                if t > maxt:
                    maxt = t
            maxt = maxt.replace(minute=0, second=0, microsecond=0) + self._hr

        mint = maxt - datetime.timedelta(days=2)

        axist = mint
        for h in range(0,49):
            x = str(240 + h*20)
            if (axist.hour == 6) or (axist.hour == 18):
                if (h>0) and (h<48):
                    self.append(tag.ClosedPart(tag_name='line', attribs={"x1":x,
                                                                         "y1":"740",
                                                                         "x2":x,
                                                                         "y2":"720",
                                                                         "stroke":self._axiscol,
                                                                         "stroke-width":"1"}))
            elif axist.hour == 12:
                if (h>0) and (h<48):
                    self.append(tag.ClosedPart(tag_name='line', attribs={"x1":x,
                                                                         "y1":"740",
                                                                         "x2":x,
                                                                         "y2":"715",
                                                                         "stroke":self._axiscol,
                                                                         "stroke-width":"1"}))
                self.append(tag.Part(tag_name='text', text="12:00", attribs={
                                                            'x':str(215 + h*20),
                                                            'y': "770",
                                                            'font-size': '20',
                                                            'font-family': self._font_family,
                                                            'fill':self._axiscol,
                                                            'stroke-width':"0"  }))
            elif axist.hour == 0:
                if (h>0) and (h<48):
                    self.append(tag.ClosedPart(tag_name='line', attribs={"x1":x,
                                                                         "y1":"740",
                                                                         "x2":x,
                                                                         "y2":"710",
                                                                         "stroke":self._axiscol,
                                                                         "stroke-width":"2"}))
                self.append(tag.Part(tag_name='text', text=axist.strftime("%d %b"), attribs={
                                                            'x':str(215 + h*20),
                                                            'y': "770",
                                                            'font-size': '20',
                                                            'font-family': self._font_family,
                                                            'fill':self._axiscol,
                                                            'stroke-width':"0"  }))
                self.append(tag.Part(tag_name='text', text="00:00", attribs={
                                                            'x':str(215 + h*20),
                                                            'y': "790",
                                                            'font-size': '20',
                                                            'font-family': self._font_family,
                                                            'fill':self._axiscol,
                                                            'stroke-width':"0"  }))

            elif (h>0) and (h<48):
                self.append(tag.ClosedPart(tag_name='line', attribs={"x1":x,
                                                                     "y1":"740",
                                                                     "x2":x,
                                                                     "y2":"730",
                                                                     "stroke":self._axiscol,
                                                                     "stroke-width":"1"}))

            axist = axist + self._hr

        # create Y axis
        minv = self.get_field_value("minvalue")
        maxv = self.get_field_value("maxvalue")

        isint = True
        try: 
            int_minv = int(minv)
        except ValueError:
            isint = False
        if isint:
            try: 
                int_maxv = int(maxv)
            except ValueError:
                isint = False

        if isint and (int_maxv > int_minv+4):
            # create a Y axis of integer values
            int_maxv, str_minv, str_maxv = self._integer_axis(int_maxv, int_minv)
        else:
            int_maxv, str_minv, str_maxv = self._float_axis(float(maxv), float(minv))

        ymin = Decimal(str_minv)
        ymax = Decimal(str_maxv)

        # y = m*val + c
        # m = 720 / (ymax-ymin)
        # c = -m*ymin

        m = Decimal("720") / (ymax-ymin)
        c = -ymin*m

        # for each point, plot a + on the graph
        for valpair in values:
            val, t = valpair
            if t < mint:
                # early point outside plottable range
                continue
            if t > maxt:
                # late point outside plottable range
                continue
            val = Decimal(val)
            if val < ymin:
                # low point outside plottable range
                continue
            if val > ymax:
                # high point outside plottable range
                continue
            tdelta = t-mint
            seconds = tdelta.total_seconds()
            x = int(960 * seconds / 172800)
            y = int(m*val + c)
            self._plot_cross(x, y)


    def _plot_cross(self, x, y):
        "plot a + on the graph at x, y"
        # Vertical line
        self.append(tag.ClosedPart(tag_name='line', attribs={"x1":str(240+x),
                                                             "y1":str(735-y),
                                                             "x2":str(240+x),
                                                             "y2":str(745-y),
                                                             "stroke":self._plotcol,
                                                             "stroke-width":"1"}))
        # Horizontal line
        self.append(tag.ClosedPart(tag_name='line', attribs={"x1":str(235+x),
                                                             "y1":str(740-y),
                                                             "x2":str(245+x),
                                                             "y2":str(740-y),
                                                             "stroke":self._plotcol,
                                                             "stroke-width":"1"}))

    def _float_axis(self, maxv, minv):
        "create a Y axis of float values, return new int_maxv, str_minv, str_maxv"
        if minv == 0.0:
            # convert 0.0045 to 45
            # convert 4500.0 to 45
            tens = math.floor(math.log10(abs(maxv)))-1
            maxy = math.ceil(maxv/10**tens)
            return self._integer_axis(maxy, 0, tens)
        else:
            diff = maxv - minv
            tens = math.floor(math.log10(abs(diff)))-1
            maxy = math.ceil(maxv/10**tens)
            miny = math.floor(minv/10**tens)
            return self._integer_axis(maxy, miny, tens)



    def _integer_axis(self, int_maxv, int_minv, tens=0):
        "create a Y axis of integer values, if necessary increase int_maxv, return new int_maxv, str_minv, str_maxv"
        # diff is the difference between minimum and maximum, divide it into intervals
        # the number of intervals should also divide 720 pixels nicely, i.e. 16, 15, 12, 10, 9, 8, 6, 5
        diff = int_maxv - int_minv

        # order of prefferred interval spacing
        prefferred = (5, 10, 2, 15, 4, 25, 20)

        difftens = math.floor(math.log10(diff))

        if difftens>2:
            prefferred = ( item * 10**(difftens-1) for item in prefferred)

        number_of_intervals = 0
        for i in prefferred:
            if (diff % i == 0) and (diff//i in (16, 15, 12, 10, 9, 8, 6, 5)):
                # intervals with spacing of i is given priority
                number_of_intervals = diff//i
                break

        if not number_of_intervals:
            if diff % 16 == 0:
                # divide axis by sixteen
                number_of_intervals = 16
            elif diff % 15 == 0:
                # divide axis by fifteen
                number_of_intervals = 15
            elif diff % 12 == 0:
                # divide axis by twelve
                number_of_intervals = 12
            elif diff % 10 == 0:
                # divide axis by ten
                number_of_intervals = 10
            elif diff % 9 == 0:
                # divide axis by nine
                number_of_intervals = 9
            elif diff % 8 == 0:
                # divide axis by eight
                number_of_intervals = 8
            elif diff % 6 == 0:
                # divide axis by six
                number_of_intervals = 6
            elif diff % 5 == 0:
                # divide axis by five
                number_of_intervals = 5
            else:
                # None of the above go nicely into diff, so add 1 to int_maxv and try again
                return self._integer_axis(int_maxv+1, int_minv, tens)

        if tens:
            mult = Decimal("1e" + str(tens))
            if abs(tens)>3:
                fstring = "0.4g"
            elif tens<0:
                fstring = "0." + str(abs(tens)) + "f"
            else:
                fstring = "0.2f"
        else:
            mult = 1
            fstring = ''

        # put the maximum value at the top of the axis
        str_maxv = format(int_maxv*mult, fstring)
        self.append(tag.Part(tag_name='text', text=str_maxv, attribs={
                                                                    'x':str(220),
                                                                    'y': "25",
                                                                    'text-anchor':'end',
                                                                    'font-size': '20',
                                                                    'font-family': self._font_family,
                                                                    'fill':self._axiscol,
                                                                    'stroke-width':"0"  }))

        # put the minimum value at the bottom of the axis
        if int_minv:
            str_minv = format(int_minv*mult, fstring)
        else:
            str_minv = "0"
        self.append(tag.Part(tag_name='text', text=str_minv, attribs={
                                                                'x':str(220),
                                                                'y': "745",
                                                                'text-anchor':'end',
                                                                'font-size': '20',
                                                                'font-family': self._font_family,
                                                                'fill':self._axiscol,
                                                                'stroke-width':"0"  }))


        # yval is the axis value at the intervals, so starting from the top
        yval = int_maxv
        y_interval = diff//number_of_intervals
        # pixel_interval is the number of pixels in the interval
        pixel_interval = 720//number_of_intervals
        # with range limits so no line at top and bottem - so rectangle not overdrawn
        for y in range(pixel_interval+20, 740, pixel_interval):
            yval -= y_interval
            self.append(tag.ClosedPart(tag_name='line', attribs={"x1":"240",
                                                                 "y1":str(y),
                                                                 "x2":"260",
                                                                 "y2":str(y),
                                                                 "stroke":self._axiscol,
                                                                 "stroke-width":"1"}))
            self.append(tag.Part(tag_name='text', text=format(yval*mult, fstring), attribs={
                                                                        'x':str(220),
                                                                        'y': str(y+5),
                                                                        'text-anchor':'end',
                                                                        'font-size': '20',
                                                                        'font-family': self._font_family,
                                                                        'fill':self._axiscol,
                                                                        'stroke-width':"0"  }))
        return int_maxv, str_minv, str_maxv



    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <rect x="240" y="20" height="720" width="960" /> <!-- the axis rectangle -->
  <!-- lines and text which draw the graph -->
</g>"""



class StarChart(Widget):

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'fill':FieldArg("text", "white"),
                        'stroke_width':FieldArg("integer", 1),
                        'stroke':FieldArg("text", "black"),
                        'stars':FieldArgTable(('text', 'text', 'text')),   # star diameter, ra, dec
                        'ra':FieldArg("text", "0"),      # right ascension 0 to 360
                        'dec':FieldArg("text", "90"),    # declination 90 to -90
                        'view':FieldArg("text", "180"),  # the field of view
                        'flip_horizontal':FieldArg("boolean", False),
                        'flip_vertical':FieldArg("boolean", False),
                        'cross':FieldArg("boolean", False),
                        'square':FieldArg("boolean", False)
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a circular star chart, held in a 500 high x 500 wide space
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self[0] = tag.ClosedPart(tag_name='circle', attribs={"cx":"250",
                                                             "cy":"250",
                                                             "r":"250",
                                                             "fill":"white",
                                                             "stroke":"black",
                                                             "stroke-width":"1"})



    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})
        stroke_width = self.get_field_value("stroke_width")
        if stroke_width:
            self[0].update_attribs({"stroke-width":str(stroke_width)})
        # stroke will be the star colour
        stroke = self.get_field_value("stroke")
        if not stroke:
            stroke = 'black'
        self[0].update_attribs({"stroke":stroke})
        fill = self.get_field_value("fill")
        if fill:
            self[0].update_attribs({"fill":fill})
        if not self.get_field_value("stars"):
            return

        # limit centre of the chart
        ra0_deg = float(self.get_field_value("ra"))
        if (ra0_deg < 0.0) or (ra0_deg > 360.0):
            ra0_deg = 0.0
        ra0 = math.radians(ra0_deg)

        dec0_deg = float(self.get_field_value("dec"))
        if dec0_deg > 90.0:
            dec0_deg = 90.0
        if dec0_deg < -90.0:
            dec0_deg = -90.0
        dec0 = math.radians(dec0_deg)

        view_deg = float(self.get_field_value("view"))

        # avoid division by zero
        if view_deg < 0.000001:
            view_deg = 0.00001

        # avoid extra wide angle
        if view_deg > 270.0:
            view_deg = 270.0

        max_dec = dec0_deg + view_deg / 2.0
        if max_dec > 90.0:
            max_dec = 90.0

        min_dec = dec0_deg - view_deg / 2.0
        if min_dec < -90.0:
            min_dec = -90.0

        scale = 500 / math.radians(view_deg)

        cosdec0 = math.cos(dec0)
        sindec0 = math.sin(dec0)

        # stereographic algorithm
        # taken from www.projectpluto.com/project.htm


        for star in self.get_field_value("stars"):

            ra_deg = float(star[1])
            dec_deg = float(star[2])

            if (ra_deg < 0.0) or (ra_deg > 360.0):
                # something wrong, do not plot this star
                continue

            # don't calculate star position if its declination is outside required view
            # unfortunately ra is more complicated
            if dec_deg > max_dec:
                continue
            if dec_deg < min_dec:
                continue

            # get the radius of a star circle to plot
            if star[0]:
                radius = float(star[0])/2.0
            else:
                radius = 0.5
            if radius < 0.1:
                radius = 0.1

            ra = math.radians(ra_deg)
            dec = math.radians(dec_deg)
            delta_ra = ra - ra0
            sindec = math.sin(dec)
            cosdec = math.cos(dec)
            cosdelta_ra = math.cos(delta_ra)

            x1 = cosdec * math.sin(delta_ra);
            y1 = sindec * cosdec0 - cosdec * cosdelta_ra * sindec0
            z1 = sindec * sindec0 + cosdec * cosdec0 * cosdelta_ra
            if z1 < -0.9:
               d = 20.0 * math.sqrt(( 1.0 - 0.81) / ( 1.00001 - z1 * z1))
            else:
               d = 2.0 / (z1 + 1.0)
            x = x1 * d * scale
            y = y1 * d * scale

            if x*x + y*y > 62500:
                # star position is outside the circle
                continue

            # move origin to circle centre (250,250)
            if self.get_field_value("flip_horizontal"):
                cx = 250 + x
            else:
                cx = 250 - x
            if self.get_field_value("flip_vertical"):
                cy = 250 + y
            else:
                cy = 250 - y
            self.append(tag.ClosedPart(tag_name='circle', attribs={"cx":str(cx),
                                                                   "cy":str(cy),
                                                                   "r":str(radius),
                                                                   "fill":stroke,
                                                                   "stroke":stroke}))


            if self.get_field_value("square"):
                # plot a square on the chart centre
                self.append(tag.ClosedPart(tag_name='rect', attribs={"x":"240",
                                                                     "y":"240",
                                                                     "width":"20",
                                                                     "height":"20",
                                                                     "style":"stroke:%s;stroke-width:1;fill-opacity:0;" % (stroke,)}))

            if self.get_field_value("cross"):
                # plot a + on the chart centre
                self.append(tag.ClosedPart(tag_name='line', attribs={"x1":"240",
                                                                     "y1":"250",
                                                                     "x2":"260",
                                                                     "y2":"250",
                                                                     "stroke":stroke,
                                                                     "stroke-width":"1"}))
                self.append(tag.ClosedPart(tag_name='line', attribs={"x1":"250",
                                                                     "y1":"240",
                                                                     "x2":"250",
                                                                     "y2":"260",
                                                                     "stroke":stroke,
                                                                     "stroke-width":"1"}))


 
    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <circle /> <!-- A circle with fill, stroke and stroke width, and diameter 500 -->
  <!-- with multiple 'star' spots, positioned according to the given ra and dec values -->
  <!-- and each with a drawn diameter given per star -->
  <!-- centre cross drawn if cross is True -->
</g>"""




class Axis1(Widget):

    # This class does not display any error messages
    display_errors = False

    _container = ((1,),)


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'fill':FieldArg("text", "white"),
                        'fill_opacity':FieldArg("text", "1"),
                        'font_family':FieldArg("text", "arial"),
                        'axiscol':FieldArg("text", "green"),
                        'minxvalue':FieldArg("text", "0"),
                        'maxxvalue':FieldArg("text", "100"),
                        'minyvalue':FieldArg("text", "0"),
                        'maxyvalue':FieldArg("text", "100")
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a graph axis, held in a 1200 high x 1200 wide space
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self._axiscol = "green"
        self._font_family = "arial"
        self[0] = tag.Part(tag_name="g")
        # The location 1 is available as a container
        self[1] = tag.Part(tag_name='g')
        self[1][0] = ''
        self._leftspace = 240
        self._topspace = 20



    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        fill = self.get_field_value("fill")
        if not fill:
            fill = "white"

        fill_opacity = self.get_field_value("fill_opacity")
        if not fill_opacity:
            fill_opacity = "1"

        self._font_family = self.get_field_value("font_family")
        if not self._font_family:
            self._font_family = "arial"

        self._axiscol = self.get_field_value("axiscol")
        if not self._axiscol:
            self._axiscol = "green"

        self[0][0] = tag.ClosedPart(tag_name='rect', attribs={"x":str(self._leftspace),
                                                           "y":str(self._topspace),
                                                           "width":"960",
                                                           "height":"720",
                                                           "fill":fill,
                                                           "fill-opacity":fill_opacity,
                                                           "stroke":self._axiscol,
                                                           "stroke-width":"1"})

        # create Y axis
        minv = self.get_field_value("minyvalue")
        maxv = self.get_field_value("maxyvalue")

        isint = True
        try: 
            int_minv = int(minv)
        except ValueError:
            isint = False
        if isint:
            try: 
                int_maxv = int(maxv)
            except ValueError:
                isint = False

        if isint and (int_maxv > int_minv+4):
            # create a Y axis of integer values
           miny,maxy = self._integer_axis(int_maxv, int_minv)
        else:
           miny,maxy = self._float_axis(float(maxv), float(minv))


        # create X axis
        minv = self.get_field_value("minxvalue")
        maxv = self.get_field_value("maxxvalue")

        isint = True
        try: 
            int_minv = int(minv)
        except ValueError:
            isint = False
        if isint:
            try: 
                int_maxv = int(maxv)
            except ValueError:
                isint = False

        if isint and (int_maxv > int_minv+4):
            # create a X axis of integer values
            minx,maxx = self._integer_axis(int_maxv, int_minv, 0, True)
        else:
            minx,maxx = self._float_axis(float(maxv), float(minv), True)

        # get line gradients and constants, note y starts from top and goes down the page
        # x = m*val + c
        # m = 960 / (xmax-xmin)
        # c = leftspace - m*xmin

        # y = m*val + c
        # m = 960 / (ymin - ymax)   - note min-max to give negative gradient
        # c = topspace - m*ymax

        my = Decimal("720") / (miny-maxy)
        cy = self._topspace - maxy*my

        mx = Decimal("960") / (maxx-minx)
        cx = self._leftspace - minx*mx 

        # ensure all contained parts have these values
        AxisLimits = namedtuple('AxisLimits', ['miny','maxy','minx','maxx', 'my', 'cy', 'mx', 'cx'])

        self.set_contained_values(AxisLimits(miny,maxy,minx,maxx,my,cy,mx,cx))



    def _float_axis(self, maxv, minv, x=False):
        "create an axis of float values"
        if minv == 0.0:
            # convert 0.0045 to 45
            # convert 4500.0 to 45
            tens = math.floor(math.log10(abs(maxv)))-1
            int_maxv = math.ceil(maxv/10**tens)
            return self._integer_axis(int_maxv, 0, tens, x)
        else:
            diff = maxv - minv
            tens = math.floor(math.log10(abs(diff)))-1
            int_maxv = math.ceil(maxv/10**tens)
            int_minv = math.floor(minv/10**tens)
            return self._integer_axis(int_maxv, int_minv, tens, x)



    def _integer_axis(self, int_maxv, int_minv, tens=0, x=False):
        "create an axis of integer values"
        # diff is the difference between minimum and maximum, divide it into intervals
        # the number of intervals should also divide 720 pixels nicely, i.e. 16, 15, 12, 10, 9, 8, 6, 5
        # or into 960 pixels, 20, 16, 15, 12, 10, 8, 6, 5
        if x:
            divlist = (20, 16, 15, 12, 10, 8, 6, 5)
        else:
            divlist = (16, 15, 12, 10, 9, 8, 6, 5)

        diff = int_maxv - int_minv

        # order of prefferred interval spacing
        prefferred = (5, 10, 2, 15, 4, 25, 20)

        difftens = math.floor(math.log10(diff))

        if difftens>2:
            prefferred = ( item * 10**(difftens-1) for item in prefferred)

        number_of_intervals = 0
        for i in prefferred:
            if (diff % i == 0) and (diff//i in divlist):
                # intervals with spacing of i is given priority
                number_of_intervals = diff//i
                break

        if not number_of_intervals:
            for divd in divlist:
                if diff % divd == 0:
                    number_of_intervals = divd
                    break
            else:
                # None of the items in divlist go nicely into diff, so add 1 to int_maxv and try again
                return self._integer_axis(int_maxv+1, int_minv, tens, x)

        if x:
            return self._x_axis(int_maxv, int_minv, tens, number_of_intervals)
        else:
            return self._y_axis(int_maxv, int_minv, tens, number_of_intervals)



    def _y_axis(self, int_maxv, int_minv, tens, number_of_intervals):
        "Draw a y axis, return decimal values of ymin,ymax"

        diff = int_maxv - int_minv

        if tens:
            mult = Decimal("1e" + str(tens))
            if abs(tens)>3:
                fstring = "0.4g"
            elif tens<0:
                fstring = "0." + str(abs(tens)) + "f"
            else:
                fstring = "0.2f"
        else:
            mult = 1
            fstring = ''

        # put the maximum value at the top of the axis
        str_maxv = format(int_maxv*mult, fstring)
        self[0].append(tag.Part(tag_name='text', text=str_maxv, attribs={
                                                                    'x':str(self._leftspace-20),
                                                                    'y': str(self._topspace+5),
                                                                    'text-anchor':'end',
                                                                    'font-size': '20',
                                                                    'font-family': self._font_family,
                                                                    'fill':self._axiscol,
                                                                    'stroke-width':"0"  }))

        # put the minimum value at the bottom of the axis
        if int_minv:
            str_minv = format(int_minv*mult, fstring)
        else:
            str_minv = "0"
        self[0].append(tag.Part(tag_name='text', text=str_minv, attribs={
                                                                'x':str(self._leftspace-20),
                                                                'y': str(720+self._topspace+5),
                                                                'text-anchor':'end',
                                                                'font-size': '20',
                                                                'font-family': self._font_family,
                                                                'fill':self._axiscol,
                                                                'stroke-width':"0"  }))


        # yval is the axis value at the intervals, so starting from the top
        yval = int_maxv
        y_interval = diff//number_of_intervals
        # pixel_interval is the number of pixels in the interval
        pixel_interval = 720//number_of_intervals
        # with range limits so no line at top and bottem - so rectangle not overdrawn
        y = self._topspace
        for n in range(int(number_of_intervals)-1):
            y += pixel_interval
            yval -= y_interval
            self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(self._leftspace-5),
                                                                 "y1":str(y),
                                                                 "x2":str(self._leftspace+20),
                                                                 "y2":str(y),
                                                                 "stroke":self._axiscol,
                                                                 "stroke-width":"1"}))
            self[0].append(tag.Part(tag_name='text', text=format(yval*mult, fstring), attribs={
                                                                        'x':str(self._leftspace-20),
                                                                        'y': str(y+5),
                                                                        'text-anchor':'end',
                                                                        'font-size': '20',
                                                                        'font-family': self._font_family,
                                                                        'fill':self._axiscol,
                                                                        'stroke-width':"0"  }))
        return Decimal(str_minv), Decimal(str_maxv)



    def _x_axis(self, int_maxv, int_minv, tens, number_of_intervals):
        "Draw a x axis, return decimal values of xmin,xmax"

        diff = int_maxv - int_minv

        if tens:
            mult = Decimal("1e" + str(tens))
            if abs(tens)>3:
                fstring = "0.4g"
            elif tens<0:
                fstring = "0." + str(abs(tens)) + "f"
            else:
                fstring = "0.2f"
        else:
            mult = 1
            fstring = ''

        # put the maximum value at the right of the axis
        str_maxv = format(int_maxv*mult, fstring)
        self[0].append(tag.Part(tag_name='text', text=str_maxv, attribs={
                                                                    'x':str(self._leftspace+960+5),
                                                                    'y': str(720+self._topspace+40),
                                                                    'text-anchor':'end',
                                                                    'font-size': '20',
                                                                    'font-family': self._font_family,
                                                                    'fill':self._axiscol,
                                                                    'stroke-width':"0"  }))

        # put the minimum value at the left of the axis
        if int_minv:
            str_minv = format(int_minv*mult, fstring)
        else:
            str_minv = "0"
        self[0].append(tag.Part(tag_name='text', text=str_minv, attribs={
                                                                'x':str(self._leftspace+5),
                                                                'y': str(720+self._topspace+40),
                                                                'text-anchor':'end',
                                                                'font-size': '20',
                                                                'font-family': self._font_family,
                                                                'fill':self._axiscol,
                                                                'stroke-width':"0"  }))


        # xval is the axis value at the intervals, so starting from the left
        xval = int_minv
        x_interval = diff//number_of_intervals
        # pixel_interval is the number of pixels in the interval
        pixel_interval = 960//number_of_intervals
        # with range limits so no line at right and left - so rectangle not overdrawn
        x = self._leftspace
        for n in range(int(number_of_intervals)-1):
            x += pixel_interval
            xval += x_interval
            self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(x),
                                                                 "y1":str(720+self._topspace+5),
                                                                 "x2":str(x),
                                                                 "y2":str(720+self._topspace-20),
                                                                 "stroke":self._axiscol,
                                                                 "stroke-width":"1"}))
            self[0].append(tag.Part(tag_name='text', text=format(xval*mult, fstring), attribs={
                                                                        'x':str(x+5),
                                                                        'y': str(720+self._topspace+40),
                                                                        'text-anchor':'end',
                                                                        'font-size': '20',
                                                                        'font-family': self._font_family,
                                                                        'fill':self._axiscol,
                                                                        'stroke-width':"0"  }))
        return Decimal(str_minv), Decimal(str_maxv)


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <g>
    <rect x="240" y="20" height="720" width="960" /> <!-- the axis rectangle -->
    <!-- lines and text which draw the axis -->
  </g>
  <g>
    <!-- Container 0 with further items, typically a widget displaying values -->
  </g>
</g>"""


class Points(Widget):

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'values':FieldArgTable(('text', 'text'), jsonset=True),
                        'pointcol':FieldArg("text", "black")
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a table of points, intended to be included in the container 0
           of an Axis widget. 
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self._minx = 0
        self._maxx = 960
        self._miny = 0
        self._maxy = 720

        self._my = -1
        self._cy = 720
        self._mx = 1
        self._cx = 0


    def set_contained_values(self, values):
        "These values set by the containing widget"
        if values.__class__.__name__ != "AxisLimits":
            return
        self._minx = values.minx
        self._maxx = values.maxx
        self._miny = values.miny
        self._maxy = values.maxy
        self._my = values.my
        self._cy = values.cy
        self._mx = values.mx
        self._cx = values.cx



    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        values = self.get_field_value("values")
        if not values:
            return

        self._pointcol = self.get_field_value("pointcol")
        if not self._pointcol:
            self._pointcol = "black"


        # for each point, plot a + on the graph
        for valpair in values:
            valx, valy = valpair
            valx = Decimal(valx)
            valy = Decimal(valy)
            if valy < self._miny:
                # low point outside plottable range
                continue
            if valy > self._maxy:
                # high point outside plottable range
                continue
            if valx < self._minx:
                # low point outside plottable range
                continue
            if valx > self._maxx:
                # high point outside plottable range
                continue
            x = int(self._mx*valx + self._cx)
            y = int(self._my*valy + self._cy)
            self._plot_cross(x, y)


    def _plot_cross(self, x, y):
        "plot a + on the graph at x, y"
        # Vertical line
        self.append(tag.ClosedPart(tag_name='line', attribs={"x1":str(x),
                                                             "y1":str(y-5),
                                                             "x2":str(x),
                                                             "y2":str(y+5),
                                                             "stroke":self._pointcol,
                                                             "stroke-width":"1"}))
        # Horizontal line
        self.append(tag.ClosedPart(tag_name='line', attribs={"x1":str(x-5),
                                                             "y1":str(y),
                                                             "x2":str(x+5),
                                                             "y2":str(y),
                                                             "stroke":self._pointcol,
                                                             "stroke-width":"1"}))


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets scale values"""
        return self._make_fieldvalues('pointcol',
                                      minx = float(self._minx),
                                      maxx = float(self._maxx),
                                      miny = float(self._miny),
                                      maxy = float(self._maxy),
                                      my = float(self._my),
                                      cy = float(self._cy),
                                      mx = float(self._mx),
                                      cx = float(self._cx))


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
    <!-- lines forming a cross at each point, positioned on the axis of the containing Axis widget -->
</g>"""



class Axis2(Widget):

    # This class does not display any error messages
    display_errors = False

    _container = ((1,),)


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'fill':FieldArg("text", "white"),
                        'fill_opacity':FieldArg("text", "1"),
                        'font_family':FieldArg("text", "arial"),
                        'axiscol':FieldArg("text", "green"),
                        'minxvalue':FieldArg("text", "0"),
                        'maxxvalue':FieldArg("text", "100"),
                        'xinterval':FieldArg("text", "20"),
                        'minyvalue':FieldArg("text", "0"),
                        'maxyvalue':FieldArg("text", "100"),
                        'yinterval':FieldArg("text", "20"),
                        'leftspace':FieldArg("text", "240"),
                        'topspace':FieldArg("text", "20"),
                        'axiswidth':FieldArg("text", "960"),
                        'axisheight':FieldArg("text", "720"),
                        'xoffset':FieldArg("boolean", False),
                        'yoffset':FieldArg("boolean", False)
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a graph axis
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self._axiscol = "green"
        self._font_family = "arial"
        self[0] = tag.Part(tag_name="g")
        # The location 1 is available as a container
        self[1] = tag.Part(tag_name='g')
        self[1][0] = ''
        self._leftspace = 240
        self._topspace = 20
        self._axiswidth = 960
        self._axisheight = 720


    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        fill = self.get_field_value("fill")
        if not fill:
            fill = "white"

        fill_opacity = self.get_field_value("fill_opacity")
        if not fill_opacity:
            fill_opacity = "0"

        self._font_family = self.get_field_value("font_family")
        if not self._font_family:
            self._font_family = "arial"

        self._axiscol = self.get_field_value("axiscol")
        if not self._axiscol:
            self._axiscol = "green"

        self._leftspace = int(self.get_field_value("leftspace"))
        self._topspace = int(self.get_field_value("topspace"))
        self._axiswidth = int(self.get_field_value("axiswidth"))
        self._axisheight = int(self.get_field_value("axisheight"))


        self[0][0] = tag.ClosedPart(tag_name='rect', attribs={"x":str(self._leftspace),
                                                           "y":str(self._topspace),
                                                           "width":str(self._axiswidth),
                                                           "height":str(self._axisheight),
                                                           "fill":fill,
                                                           "fill-opacity":fill_opacity,
                                                           "stroke":self._axiscol,
                                                           "stroke-width":"1"})

        # create Y axis
        minv = Decimal(self.get_field_value("minyvalue"))
        maxv = Decimal(self.get_field_value("maxyvalue"))
        interval = Decimal(self.get_field_value("yinterval"))
        if (maxv <= minv):
            return
        if (maxv-minv)<interval:
            return
 
        miny,maxy = self._y_axis(maxv, minv, interval)


        # create X axis
        minv = Decimal(self.get_field_value("minxvalue"))
        maxv = Decimal(self.get_field_value("maxxvalue"))
        interval = Decimal(self.get_field_value("xinterval"))
        if (maxv <= minv):
            return
        if (maxv-minv)<interval:
            return

        minx,maxx = self._x_axis(maxv, minv, interval)

        # get line gradients and constants, note y starts from top and goes down the page
        # x = m*val + c
        # m = 960 / (xmax-xmin)
        # c = leftspace - m*xmin

        # y = m*val + c
        # m = 960 / (ymin - ymax)   - note min-max to give negative gradient
        # c = topspace - m*ymax

        my = Decimal(str(self._axisheight)) / (miny-maxy)
        cy = self._topspace - maxy*my

        mx = Decimal(str(self._axiswidth)) / (maxx-minx)
        cx = self._leftspace - minx*mx

        # ensure all contained parts have these values
        AxisLimits = namedtuple('AxisLimits', ['miny','maxy','minx','maxx', 'my', 'cy', 'mx', 'cx'])

        self.set_contained_values(AxisLimits(miny,maxy,minx,maxx,my,cy,mx,cx))




    def _y_axis(self, maxv, minv, interval):
        "Draw a y axis, return decimal values of ymin,ymax"

        fraction_number_of_intervals = (maxv-minv)/interval
        number_of_intervals = fraction_number_of_intervals.to_integral_value(rounding=ROUND_UP)
        maxv = minv + number_of_intervals*interval


        yoffset = self.get_field_value("yoffset")
        if yoffset:
            maxv += interval/Decimal("2.0")
            minv -= interval/Decimal("2.0")
            # pixel_interval is the number of pixels in the interval - but in this case there is an extra interval
            pixel_interval = int(self._axisheight//(number_of_intervals+1))
        else:
            # pixel_interval is the number of pixels in the interval
            pixel_interval = int(self._axisheight//number_of_intervals)


        # yval is the axis value at the intervals, so starting from the top
        # y is the pixel value to be plotted
        if yoffset:
            y = self._topspace - pixel_interval//2
            yval = maxv + interval/Decimal("2.0")
            for n in range(int(number_of_intervals)+1):
                y += pixel_interval
                yval -= interval
                self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(self._leftspace-5),
                                                                     "y1":str(y),
                                                                     "x2":str(self._leftspace+20),
                                                                     "y2":str(y),
                                                                     "stroke":self._axiscol,
                                                                     "stroke-width":"1"}))
                self[0].append(tag.Part(tag_name='text', text=str(yval), attribs={
                                                                            'x':str(self._leftspace-20),
                                                                            'y': str(y+5),
                                                                            'text-anchor':'end',
                                                                            'font-size': '20',
                                                                            'font-family': self._font_family,
                                                                            'fill':self._axiscol,
                                                                            'stroke-width':"0"  }))
        else:
            y = self._topspace
            yval = maxv
            for n in range(int(number_of_intervals)-1):
                y += pixel_interval
                yval -= interval
                self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(self._leftspace-5),
                                                                     "y1":str(y),
                                                                     "x2":str(self._leftspace+20),
                                                                     "y2":str(y),
                                                                     "stroke":self._axiscol,
                                                                     "stroke-width":"1"}))
                self[0].append(tag.Part(tag_name='text', text=str(yval), attribs={
                                                                            'x':str(self._leftspace-20),
                                                                            'y': str(y+5),
                                                                            'text-anchor':'end',
                                                                            'font-size': '20',
                                                                            'font-family': self._font_family,
                                                                            'fill':self._axiscol,
                                                                            'stroke-width':"0"  }))

            # put the maximum value at the top of the axis
            self[0].append(tag.Part(tag_name='text', text=str(maxv), attribs={
                                                                        'x':str(self._leftspace-20),
                                                                        'y': str(self._topspace+5),
                                                                        'text-anchor':'end',
                                                                        'font-size': '20',
                                                                        'font-family': self._font_family,
                                                                        'fill':self._axiscol,
                                                                        'stroke-width':"0"  }))

            # put the minimum value at the bottom of the axis
            self[0].append(tag.Part(tag_name='text', text=str(minv), attribs={
                                                                    'x':str(self._leftspace-20),
                                                                    'y': str(self._axisheight+self._topspace+5),
                                                                    'text-anchor':'end',
                                                                    'font-size': '20',
                                                                    'font-family': self._font_family,
                                                                    'fill':self._axiscol,
                                                                    'stroke-width':"0"  }))

        return minv,maxv



    def _x_axis(self, maxv, minv, interval):
        "Draw a x axis, return decimal values of xmin,xmax"

        fraction_number_of_intervals = (maxv-minv)/interval
        number_of_intervals = fraction_number_of_intervals.to_integral_value(rounding=ROUND_UP)
        maxv = minv + number_of_intervals*interval


        xoffset = self.get_field_value("xoffset")
        if xoffset:
            maxv += interval/Decimal("2.0")
            minv -= interval/Decimal("2.0")
            # pixel_interval is the number of pixels in the interval - but in this case there is an extra interval
            pixel_interval = int(self._axiswidth//(number_of_intervals+1))
        else:
            # pixel_interval is the number of pixels in the interval
            pixel_interval = int(self._axiswidth//number_of_intervals)

        # xval is the axis value at the intervals, so starting from the left
        # x is the pixel value to be plotted
        if xoffset:
            x = self._leftspace - pixel_interval//2
            xval = minv - interval/Decimal("2.0")
            for n in range(int(number_of_intervals)+1):
                x += pixel_interval
                xval += interval
                self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(x),
                                                                     "y1":str(self._axisheight+self._topspace+5),
                                                                     "x2":str(x),
                                                                     "y2":str(self._axisheight+self._topspace-20),
                                                                     "stroke":self._axiscol,
                                                                     "stroke-width":"1"}))
                self[0].append(tag.Part(tag_name='text', text=str(xval), attribs={
                                                                            'x':str(x+5),
                                                                            'y': str(self._axisheight+self._topspace+40),
                                                                            'text-anchor':'end',
                                                                            'font-size': '20',
                                                                            'font-family': self._font_family,
                                                                            'fill':self._axiscol,
                                                                            'stroke-width':"0"  }))
        else:
            x = self._leftspace
            xval = minv
            for n in range(int(number_of_intervals)-1):
                x += pixel_interval
                xval += interval
                self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(x),
                                                                     "y1":str(self._axisheight+self._topspace+5),
                                                                     "x2":str(x),
                                                                     "y2":str(self._axisheight+self._topspace-20),
                                                                     "stroke":self._axiscol,
                                                                     "stroke-width":"1"}))
                self[0].append(tag.Part(tag_name='text', text=str(xval), attribs={
                                                                            'x':str(x+5),
                                                                            'y': str(self._axisheight+self._topspace+40),
                                                                            'text-anchor':'end',
                                                                            'font-size': '20',
                                                                            'font-family': self._font_family,
                                                                            'fill':self._axiscol,
                                                                            'stroke-width':"0"  }))
      
            # put the maximum value at the right of the axis
            self[0].append(tag.Part(tag_name='text', text=str(maxv), attribs={
                                                                        'x':str(self._leftspace+self._axiswidth+5),
                                                                        'y': str(self._axisheight+self._topspace+40),
                                                                        'text-anchor':'end',
                                                                        'font-size': '20',
                                                                        'font-family': self._font_family,
                                                                        'fill':self._axiscol,
                                                                        'stroke-width':"0"  }))

            # put the minimum value at the left of the axis
            self[0].append(tag.Part(tag_name='text', text=str(minv), attribs={
                                                                    'x':str(self._leftspace+5),
                                                                    'y': str(self._axisheight+self._topspace+40),
                                                                    'text-anchor':'end',
                                                                    'font-size': '20',
                                                                    'font-family': self._font_family,
                                                                    'fill':self._axiscol,
                                                                    'stroke-width':"0"  }))

        return minv, maxv


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <g>
    <!-- rectangle with given position, height and width -->
    <!-- lines and text which draw the axis -->
  </g>
  <g>
    <!-- Container 0 with further items, typically a widget displaying values -->
  </g>
</g>"""



class Lines(Widget):

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'values':FieldArgTable(('text', 'text'), jsonset=True),
                        'linecol':FieldArg("text", "black"),
                        'linewidth':FieldArg("text", "2"),
                        'pointradius':FieldArg("text", "1")
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a table of points joined by lines, intended to be included in the container 0
           of an Axis widget. 
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self._minx = 0
        self._maxx = 960
        self._miny = 0
        self._maxy = 720

        self._my = -1
        self._cy = 720
        self._mx = 1
        self._cx = 0


    def set_contained_values(self, values):
        "These values set by the containing widget"
        if values.__class__.__name__ != "AxisLimits":
            return
        self._minx = values.minx
        self._maxx = values.maxx
        self._miny = values.miny
        self._maxy = values.maxy
        self._my = values.my
        self._cy = values.cy
        self._mx = values.mx
        self._cx = values.cx


    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        values = self.get_field_value("values")
        if not values:
            return

        self._linecol = self.get_field_value("linecol")
        if not self._linecol:
            self._linecol = "black"

        self._linewidth = self.get_field_value("linewidth")

        self._pointradius = self.get_field_value("pointradius")
        if not self._pointradius:
            self._pointradius = "2"

        old_x = ''
        old_y = ''
        # for each point, plot a circle on the graph
        for valpair in values:
            valx, valy = valpair
            valx = Decimal(valx)
            valy = Decimal(valy)
            if valy < self._miny:
                # low point outside plottable range
                continue
            if valy > self._maxy:
                # high point outside plottable range
                continue
            if valx < self._minx:
                # low point outside plottable range
                continue
            if valx > self._maxx:
                # high point outside plottable range
                continue
            x = int(self._mx*valx + self._cx)
            y = int(self._my*valy + self._cy)
            # plot a circle at point
            self.append(tag.ClosedPart(tag_name='circle', attribs={"cx":str(x),
                                                                   "cy":str(y),
                                                                   "r":str(self._pointradius),
                                                                   "fill":self._linecol,
                                                                   "stroke":self._linecol}))
            # plot a line from previous point
            if old_x and self._linewidth:
                self.append(tag.ClosedPart(tag_name='line', attribs={"x1":old_x,
                                                                     "y1":old_y,
                                                                     "x2":str(x),
                                                                     "y2":str(y),
                                                                     "stroke":self._linecol,
                                                                     "stroke-width":self._linewidth}))
            old_x = str(x)
            old_y = str(y)


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets scale values"""
        return self._make_fieldvalues('linecol',
                                      'linewidth',
                                      'pointradius',
                                      minx = float(self._minx),
                                      maxx = float(self._maxx),
                                      miny = float(self._miny),
                                      maxy = float(self._maxy),
                                      my = float(self._my),
                                      cy = float(self._cy),
                                      mx = float(self._mx),
                                      cx = float(self._cx))


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
    <!-- lines linking value points, positioned on the axis of the containing Axis widget -->
</g>"""



class XBars(Widget):

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'values':FieldArgTable(('text', 'text'), jsonset=True),
                        'bar_width':FieldArg("text", "10"),
                        'fill':FieldArg("text", "green"),
                        'fill_opacity':FieldArg("text", "1"),
                        'stroke':FieldArg("text", "black"),
                        'stroke_width':FieldArg("text", "1")
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a series of vertical bars, intended to be included in the container 0
           of an Axis widget. 
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self._minx = 0
        self._maxx = 960
        self._miny = 0
        self._maxy = 720
        self._my = -1
        self._cy = 720
        self._mx = 1
        self._cx = 0
        self._barwidth = 10
        self._yaxis = 720


    def set_contained_values(self, values):
        "These values set by the containing widget"
        if values.__class__.__name__ != "AxisLimits":
            return
        self._minx = values.minx
        self._maxx = values.maxx
        self._miny = values.miny
        self._maxy = values.maxy
        self._my = values.my
        self._cy = values.cy
        self._mx = values.mx
        self._cx = values.cx



    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        values = self.get_field_value("values")
        if not values:
            return

        bar_width = self.get_field_value("bar_width")
        if not bar_width:
            return

        # get bar in pixels
        self._barwidth = int(self._mx*Decimal(bar_width))
        if not self._barwidth:
            return

        self._halfbar = self._barwidth//2
        self._yaxis = int(self._my*self._miny + self._cy)

        self._fill = self.get_field_value("fill")
        self._fill_opacity = self.get_field_value("fill_opacity")
        self._stroke = self.get_field_value("stroke")
        self._stroke_width = self.get_field_value("stroke_width")

        # for each point, plot a bar on the graph
        for valpair in values:
            valx, valy = valpair
            valx = Decimal(valx)
            valy = Decimal(valy)
            if valy < self._miny:
                # low point outside plottable range
                continue
            if valy > self._maxy:
                # high point outside plottable range
                continue
            if valx < self._minx:
                # low point outside plottable range
                continue
            if valx > self._maxx:
                # high point outside plottable range
                continue
            x = int(self._mx*valx + self._cx)
            y = int(self._my*valy + self._cy)
            self._plot_bar(x, y)


    def _plot_bar(self, x, y):
        "plot a bar on the graph at x, y"
        self.append(tag.ClosedPart(tag_name='rect', attribs={"x":str(x-self._halfbar),
                                                            "y":str(y),
                                                            "width":str(self._barwidth),
                                                            "height":str(self._yaxis-y),
                                                            "fill":self._fill,
                                                            "fill-opacity":self._fill_opacity,
                                                            "stroke":self._stroke,
                                                            "stroke-width":self._stroke_width}))



    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets scale values"""
        return self._make_fieldvalues(
                                     'fill',
                                     'fill_opacity',
                                     'stroke',
                                     'stroke_width',
                                      barwidth = self._barwidth,
                                      yaxis = self._yaxis,
                                      minx = float(self._minx),
                                      maxx = float(self._maxx),
                                      miny = float(self._miny),
                                      maxy = float(self._maxy),
                                      my = float(self._my),
                                      cy = float(self._cy),
                                      mx = float(self._mx),
                                      cx = float(self._cx))


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
    <!-- vertical rectangle bars at each X point, positioned on the axis of the containing Axis widget -->
</g>"""


class YBars(Widget):

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'values':FieldArgTable(('text', 'text'), jsonset=True),
                        'bar_width':FieldArg("text", "10"),
                        'fill':FieldArg("text", "green"),
                        'fill_opacity':FieldArg("text", "1"),
                        'stroke':FieldArg("text", "black"),
                        'stroke_width':FieldArg("text", "1")
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a series of horizontal bars, intended to be included in the container 0
           of an Axis widget. 
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self._minx = 0
        self._maxx = 960
        self._miny = 0
        self._maxy = 720
        self._my = -1
        self._cy = 720
        self._mx = 1
        self._cx = 0
        self._barwidth = 10
        self._xaxis = 960


    def set_contained_values(self, values):
        "These values set by the containing widget"
        if values.__class__.__name__ != "AxisLimits":
            return
        self._minx = values.minx
        self._maxx = values.maxx
        self._miny = values.miny
        self._maxy = values.maxy
        self._my = values.my
        self._cy = values.cy
        self._mx = values.mx
        self._cx = values.cx



    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        values = self.get_field_value("values")
        if not values:
            return

        bar_width = self.get_field_value("bar_width")
        if not bar_width:
            return

        # get bar in pixels, -1 multiplier as my is negative
        self._barwidth = int(-1*self._my*Decimal(bar_width))
        if not self._barwidth:
            return

        self._halfbar = self._barwidth//2
        self._xaxis = int(self._mx*self._minx + self._cx)

        self._fill = self.get_field_value("fill")
        self._fill_opacity = self.get_field_value("fill_opacity")
        self._stroke = self.get_field_value("stroke")
        self._stroke_width = self.get_field_value("stroke_width")

        # for each point, plot a bar on the graph
        for valpair in values:
            valx, valy = valpair
            valx = Decimal(valx)
            valy = Decimal(valy)
            if valy < self._miny:
                # low point outside plottable range
                continue
            if valy > self._maxy:
                # high point outside plottable range
                continue
            if valx < self._minx:
                # low point outside plottable range
                continue
            if valx > self._maxx:
                # high point outside plottable range
                continue
            x = int(self._mx*valx + self._cx)
            y = int(self._my*valy + self._cy)
            self._plot_bar(x, y)


    def _plot_bar(self, x, y):
        "plot a bar on the graph at x, y"
        self.append(tag.ClosedPart(tag_name='rect', attribs={"x":str(self._xaxis),
                                                            "y":str(y-self._halfbar),
                                                            "width":str(x-self._xaxis),
                                                            "height":str(self._barwidth),
                                                            "fill":self._fill,
                                                            "fill-opacity":self._fill_opacity,
                                                            "stroke":self._stroke,
                                                            "stroke-width":self._stroke_width}))


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets scale values"""
        return self._make_fieldvalues(
                                     'fill',
                                     'fill_opacity',
                                     'stroke',
                                     'stroke_width',
                                      barwidth = self._barwidth,
                                      xaxis = self._xaxis,
                                      minx = float(self._minx),
                                      maxx = float(self._maxx),
                                      miny = float(self._miny),
                                      maxy = float(self._maxy),
                                      my = float(self._my),
                                      cy = float(self._cy),
                                      mx = float(self._mx),
                                      cx = float(self._cx))


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
    <!-- horizontal rectangle bars at each Y point, positioned on the axis of the containing Axis widget -->
</g>"""


class Axis3(Widget):

    # This class does not display any error messages
    display_errors = False

    _container = ((1,),)


    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True),
                        'fill':FieldArg("text", "white"),
                        'fill_opacity':FieldArg("text", "1"),
                        'font_family':FieldArg("text", "arial"),
                        'axiscol':FieldArg("text", "green"),
                        'minxvalue':FieldArg("text", "0"),
                        'maxxvalue':FieldArg("text", "100"),
                        'xlabels':FieldArgList('text'),
                        'ylabels':FieldArgList('text'),
                        'minyvalue':FieldArg("text", "0"),
                        'maxyvalue':FieldArg("text", "100"),
                        'leftspace':FieldArg("text", "240"),
                        'topspace':FieldArg("text", "20"),
                        'axiswidth':FieldArg("text", "960"),
                        'axisheight':FieldArg("text", "720"),
                        'xoffset':FieldArg("boolean", False),
                        'yoffset':FieldArg("boolean", False)
                       }


    def __init__(self, name=None, brief='', **field_args):
        """A g element which holds a graph axis
        """
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self._axiscol = "green"
        self._font_family = "arial"
        self[0] = tag.Part(tag_name="g")
        # The location 1 is available as a container
        self[1] = tag.Part(tag_name='g')
        self[1][0] = ''
        self._leftspace = 240
        self._topspace = 20
        self._axiswidth = 960
        self._axisheight = 720


    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

        fill = self.get_field_value("fill")
        if not fill:
            fill = "white"

        fill_opacity = self.get_field_value("fill_opacity")
        if not fill_opacity:
            fill_opacity = "0"

        self._font_family = self.get_field_value("font_family")
        if not self._font_family:
            self._font_family = "arial"

        self._axiscol = self.get_field_value("axiscol")
        if not self._axiscol:
            self._axiscol = "green"

        self._leftspace = int(self.get_field_value("leftspace"))
        self._topspace = int(self.get_field_value("topspace"))
        self._axiswidth = int(self.get_field_value("axiswidth"))
        self._axisheight = int(self.get_field_value("axisheight"))


        self[0][0] = tag.ClosedPart(tag_name='rect', attribs={"x":str(self._leftspace),
                                                           "y":str(self._topspace),
                                                           "width":str(self._axiswidth),
                                                           "height":str(self._axisheight),
                                                           "fill":fill,
                                                           "fill-opacity":fill_opacity,
                                                           "stroke":self._axiscol,
                                                           "stroke-width":"1"})

        xlabels = self.get_field_value('xlabels')
        ylabels = self.get_field_value('ylabels')


        # create Y axis
        minv = Decimal(self.get_field_value("minyvalue"))
        maxv = Decimal(self.get_field_value("maxyvalue"))
        if (maxv <= minv):
            return
        if ylabels and (len(ylabels) > 1): 
            interval = (maxv-minv)/(len(ylabels)-1)
        else:
            interval = (maxv-minv)/10
 
        miny,maxy = self._y_axis(maxv, minv, interval, ylabels)


        # create X axis
        minv = Decimal(self.get_field_value("minxvalue"))
        maxv = Decimal(self.get_field_value("maxxvalue"))
        if (maxv <= minv):
            return
        if xlabels and (len(xlabels) > 1): 
            interval = (maxv-minv)/(len(xlabels)-1)
        else:
            interval = (maxv-minv)/10

        minx,maxx = self._x_axis(maxv, minv, interval, xlabels)

        # get line gradients and constants, note y starts from top and goes down the page
        # x = m*val + c
        # m = 960 / (xmax-xmin)
        # c = leftspace - m*xmin

        # y = m*val + c
        # m = 960 / (ymin - ymax)   - note min-max to give negative gradient
        # c = topspace - m*ymax

        my = Decimal(str(self._axisheight)) / (miny-maxy)
        cy = self._topspace - maxy*my

        mx = Decimal(str(self._axiswidth)) / (maxx-minx)
        cx = self._leftspace - minx*mx

        # ensure all contained parts have these values
        AxisLimits = namedtuple('AxisLimits', ['miny','maxy','minx','maxx', 'my', 'cy', 'mx', 'cx'])

        self.set_contained_values(AxisLimits(miny,maxy,minx,maxx,my,cy,mx,cx))




    def _y_axis(self, maxv, minv, interval, labels):
        "Draw a y axis, return decimal values of ymin,ymax"

        fraction_number_of_intervals = (maxv-minv)/interval
        number_of_intervals = fraction_number_of_intervals.to_integral_value(rounding=ROUND_UP)
        maxv = minv + number_of_intervals*interval


        yoffset = self.get_field_value("yoffset")
        if yoffset:
            maxv += interval/Decimal("2.0")
            minv -= interval/Decimal("2.0")
            # pixel_interval is the number of pixels in the interval - but in this case there is an extra interval
            pixel_interval = int(self._axisheight//(number_of_intervals+1))
        else:
            # pixel_interval is the number of pixels in the interval
            pixel_interval = int(self._axisheight//number_of_intervals)


        # yval is the axis value at the intervals, so starting from the top
        # y is the pixel value to be plotted
        if yoffset:
            y = self._topspace - pixel_interval//2
            yval = maxv + interval/Decimal("2.0")
            lbl = int(number_of_intervals)
            for n in range(int(number_of_intervals)+1):
                y += pixel_interval
                yval -= interval
                self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(self._leftspace-5),
                                                                     "y1":str(y),
                                                                     "x2":str(self._leftspace+20),
                                                                     "y2":str(y),
                                                                     "stroke":self._axiscol,
                                                                     "stroke-width":"1"}))
                if labels:
                    try:
                        ylabel = labels[lbl]
                    except IndexError as e:
                        ylabel = str(yval)
                    lbl -= 1
                else:
                    ylabel = str(yval)

                self[0].append(tag.Part(tag_name='text', text=ylabel, attribs={
                                                                            'x':str(self._leftspace-20),
                                                                            'y': str(y+5),
                                                                            'text-anchor':'end',
                                                                            'font-size': '20',
                                                                            'font-family': self._font_family,
                                                                            'fill':self._axiscol,
                                                                            'stroke-width':"0"  }))
        else:
            y = self._topspace
            yval = maxv
            lbl = int(number_of_intervals)-1
            for n in range(int(number_of_intervals)-1):
                y += pixel_interval
                yval -= interval
                self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(self._leftspace-5),
                                                                     "y1":str(y),
                                                                     "x2":str(self._leftspace+20),
                                                                     "y2":str(y),
                                                                     "stroke":self._axiscol,
                                                                     "stroke-width":"1"}))
                if labels:
                    try:
                        ylabel = labels[lbl]
                    except IndexError as e:
                        ylabel = str(yval)
                    lbl -= 1
                else:
                    ylabel = str(yval)
                self[0].append(tag.Part(tag_name='text', text=ylabel, attribs={
                                                                            'x':str(self._leftspace-20),
                                                                            'y': str(y+5),
                                                                            'text-anchor':'end',
                                                                            'font-size': '20',
                                                                            'font-family': self._font_family,
                                                                            'fill':self._axiscol,
                                                                            'stroke-width':"0"  }))

            # put the maximum value at the top of the axis
            if labels:
                try:
                    ylabel = labels[int(number_of_intervals)]
                except IndexError as e:
                    ylabel = str(maxv)
            else:
                ylabel = str(maxv)
            self[0].append(tag.Part(tag_name='text', text=ylabel, attribs={
                                                                        'x':str(self._leftspace-20),
                                                                        'y': str(self._topspace+5),
                                                                        'text-anchor':'end',
                                                                        'font-size': '20',
                                                                        'font-family': self._font_family,
                                                                        'fill':self._axiscol,
                                                                        'stroke-width':"0"  }))

            # put the minimum value at the bottom of the axis
            if labels:
                try:
                    ylabel = labels[0]
                except IndexError as e:
                    ylabel = str(minv)
            else:
                ylabel = str(minv)
            self[0].append(tag.Part(tag_name='text', text=ylabel, attribs={
                                                                    'x':str(self._leftspace-20),
                                                                    'y': str(self._axisheight+self._topspace+5),
                                                                    'text-anchor':'end',
                                                                    'font-size': '20',
                                                                    'font-family': self._font_family,
                                                                    'fill':self._axiscol,
                                                                    'stroke-width':"0"  }))

        return minv,maxv



    def _x_axis(self, maxv, minv, interval, labels):
        "Draw a x axis, return decimal values of xmin,xmax"

        fraction_number_of_intervals = (maxv-minv)/interval
        number_of_intervals = fraction_number_of_intervals.to_integral_value(rounding=ROUND_UP)
        maxv = minv + number_of_intervals*interval


        xoffset = self.get_field_value("xoffset")
        if xoffset:
            maxv += interval/Decimal("2.0")
            minv -= interval/Decimal("2.0")
            # pixel_interval is the number of pixels in the interval - but in this case there is an extra interval
            pixel_interval = int(self._axiswidth//(number_of_intervals+1))
        else:
            # pixel_interval is the number of pixels in the interval
            pixel_interval = int(self._axiswidth//number_of_intervals)

        # xval is the axis value at the intervals, so starting from the left
        # x is the pixel value to be plotted
        if xoffset:
            x = self._leftspace - pixel_interval//2
            xval = minv - interval/Decimal("2.0")
            lbl = 0
            for n in range(int(number_of_intervals)+1):
                x += pixel_interval
                xval += interval
                self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(x),
                                                                     "y1":str(self._axisheight+self._topspace+5),
                                                                     "x2":str(x),
                                                                     "y2":str(self._axisheight+self._topspace-20),
                                                                     "stroke":self._axiscol,
                                                                     "stroke-width":"1"}))
                if labels:
                    try:
                        xlabel = labels[lbl]
                    except IndexError as e:
                        xlabel = str(xval)
                    lbl += 1
                else:
                    xlabel = str(xval)

                self[0].append(tag.Part(tag_name='text', text=xlabel, attribs={
                                                                            'x':str(x+5),
                                                                            'y': str(self._axisheight+self._topspace+40),
                                                                            'text-anchor':'end',
                                                                            'font-size': '20',
                                                                            'font-family': self._font_family,
                                                                            'fill':self._axiscol,
                                                                            'stroke-width':"0"  }))
        else:
            x = self._leftspace
            xval = minv
            lbl = 1
            for n in range(int(number_of_intervals)-1):
                x += pixel_interval
                xval += interval
                self[0].append(tag.ClosedPart(tag_name='line', attribs={"x1":str(x),
                                                                     "y1":str(self._axisheight+self._topspace+5),
                                                                     "x2":str(x),
                                                                     "y2":str(self._axisheight+self._topspace-20),
                                                                     "stroke":self._axiscol,
                                                                     "stroke-width":"1"}))
                if labels:
                    try:
                        xlabel = labels[lbl]
                    except IndexError as e:
                        xlabel = str(xval)
                    lbl += 1
                else:
                    xlabel = str(xval)
                self[0].append(tag.Part(tag_name='text', text=xlabel, attribs={
                                                                            'x':str(x+5),
                                                                            'y': str(self._axisheight+self._topspace+40),
                                                                            'text-anchor':'end',
                                                                            'font-size': '20',
                                                                            'font-family': self._font_family,
                                                                            'fill':self._axiscol,
                                                                            'stroke-width':"0"  }))
      
            # put the maximum value at the right of the axis
            if labels:
                try:
                    xlabel = labels[lbl]
                except IndexError as e:
                    xlabel = str(maxv)
            else:
                xlabel = str(maxv)
            self[0].append(tag.Part(tag_name='text', text=xlabel, attribs={
                                                                        'x':str(self._leftspace+self._axiswidth+5),
                                                                        'y': str(self._axisheight+self._topspace+40),
                                                                        'text-anchor':'end',
                                                                        'font-size': '20',
                                                                        'font-family': self._font_family,
                                                                        'fill':self._axiscol,
                                                                        'stroke-width':"0"  }))

            # put the minimum value at the left of the axis
            if labels:
                try:
                    xlabel = labels[0]
                except IndexError as e:
                    xlabel = str(minv)
            else:
                xlabel = str(minv)
            self[0].append(tag.Part(tag_name='text', text=xlabel, attribs={
                                                                    'x':str(self._leftspace+5),
                                                                    'y': str(self._axisheight+self._topspace+40),
                                                                    'text-anchor':'end',
                                                                    'font-size': '20',
                                                                    'font-family': self._font_family,
                                                                    'fill':self._axiscol,
                                                                    'stroke-width':"0"  }))

        return minv, maxv


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <g>
    <!-- rectangle with given position, height and width -->
    <!-- lines and labels which draw the axis -->
  </g>
  <g>
    <!-- Container 0 with further items, typically a widget displaying values -->
  </g>
</g>"""








"""Functions for editing a page"""

from collections import OrderedDict

from ..ski import skiboot, tag, read_json, dump_project
from ..ski.excepts import ServerError

from . import project_loaded, item_info, get_proj_page, del_location_in_page, insert_item_in_page

from .info_tuple import PageElement, PageTextBlock, WidgFieldInfo


def pagechange(project, pagenumber):
    """Returns the page change uuid

       Given a project name, and the integer ident number of a page within the project,
       it returns a 'change' uuid associated  with the page. A new random uuid is generated
       every time a page is changed, and is therefore a measure that a given page both exists,
       and has not changed since the last time it was accessed.
       The project must be currently loaded as either the root project or a sub-project.
       Returns None if no page with this pagenumber is found."""
    # raise error if invalid project
    project_loaded(project)
    info = item_info(project, pagenumber)
    if not info:
        return
    if info.item_type == "Folder":
        return
    return info.change


def rename_page(project, pagenumber, pchange, newname):
    """Rename the given page.

       Given a project name, the integer ident number of a page within the project,
       the page change uuid and a new page name, it changes the page name and returns
       a new uuid on success, or raises a ServerError on failure.
       The project must be currently loaded as either the root project or a sub-project."""
    if not newname:
        raise ServerError(message="No new page name given")
    # get a copy of the page, which can have a new name set
    # and can then be saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    # Rename the page
    page.name = newname
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def page_description(project, pagenumber, pchange, brief):
    "Set a page brief description"
    if not brief:
        return "No new page name given"
    # get a copy of the page, which can have a new brief set
    # and can then be saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    # Set the page brief
    page.brief = brief
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def page_string(project, pagenumber):
    "Returns a string of the page, may be usefull for diagnostics"
    proj, page = get_proj_page(project, pagenumber)
    return str(page)


def page_language(project, pagenumber, pchange, language):
    "Sets language in the page html tag"
    # get a copy of the page, which can have a new language set
    # and can then be saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "TemplatePage":
        raise ServerError(message = "Invalid page type")
    # Set the page language
    page.lang = language
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def get_page_backcol(project, pagenumber, pchange):
    """Returns tuple, first element True if the background colour is set in the html tag
                      second element being the background colour string"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "TemplatePage":
        raise ServerError(message = "Invalid page type")
    return page.show_backcol, page.backcol


def page_backcol(project, pagenumber, pchange, show_backcol, backcol):
    """Sets the backcolour and enables/disables it in the html tag"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "TemplatePage":
        raise ServerError(message = "Invalid page type")
    page.show_backcol = show_backcol
    page.backcol = backcol
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def page_last_scroll(project, pagenumber, pchange, last_scroll):
    "Sets last_scroll flag in the page"
    # get a copy of the page, which can be edited
    # and then saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "TemplatePage":
        raise ServerError(message = "Invalid page type")
    # Set the page last_scroll
    page.last_scroll = last_scroll
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def get_page_interval(project, pagenumber, pchange):
    """Returns tuple, first element is the refresh interval, or 0 if target does not exist
                      second element being the interval target, or None if not set"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "TemplatePage":
        raise ServerError(message = "Invalid page type")
    if page.interval != 0:
        if page.interval_target is None:
            return 0, None
        interval_target = page.interval_target
        # if interval target is an ident of this project, only return an integer
        if isinstance(interval_target, skiboot.Ident):
            if interval_target.proj == project:
                interval_target = str(interval_target.num)
            else:
                # ident is another project, put the full ident
                interval_target = interval_target.to_comma_str()
        return page.interval, interval_target
    else:
        return 0, None


def page_interval(project, pagenumber, pchange, interval, interval_target):
    """Sets the refresh interval and target"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "TemplatePage":
        raise ServerError(message = "Invalid page type")
    interval_target = skiboot.make_ident_or_label(interval_target, project)
    # set page attributes
    if (interval == 0) or (not interval_target):
        page.interval = 0
        page.interval_target = None
    else:
        page.interval = interval
        page.interval_target = interval_target
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def page_default_error_widget(project, pagenumber, pchange, default_error_widget):
    "Sets default_error_widget in the page"
    # get a copy of the page, which can be edited
    # and then saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "TemplatePage":
        raise ServerError(message = "Invalid page type")
    page.default_error_widget = default_error_widget
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def page_enable_cache(project, pagenumber, pchange, enable_cache):
    "Sets enable cache in the page"
    # get a copy of the page, which can be edited
    # and then saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if (page.page_type != "SVG") and (page.page_type != "JSON") and (page.page_type != "FilePage") and (page.page_type != "CSS"):
        raise ServerError(message = "Invalid page type")
    page.enable_cache = enable_cache
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def new_parent(project, pagenumber, pchange, new_parent_number):
    "Gives a page a new parent folder"

    # get a copy of the page, which is to have a new parent folder set
    # and can then be saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if not isinstance(new_parent_number, int):
        raise ServerError(message="Given new_parent_number is not an integer")
    ident = skiboot.Ident.to_ident((project, new_parent_number))
    if ident is None:
        raise ServerError(message="Invalid project, new_parent_number")
    folder = skiboot.from_ident(ident, project)
    if folder is None:
        raise ServerError(message="Invalid Folder")
    if folder.page_type != 'Folder':
        raise ServerError(message="Item is not a page")
    old_parent = page.parentfolder
    if old_parent == folder:
        raise ServerError(message="Parent folder unchanged?")
    if page.name in folder:
        raise ServerError(message="The folder already contains an item with this name")
    return proj.save_page(page, folder.ident)


def delete_page(project, pagenumber):
    "delete this page, return None on success, error message on failure"
    try:
        proj, page = get_proj_page(project, pagenumber)
        # delete the page from the project
        proj.delete_item(page.ident)
    except ServerError as e:
        return e.message



def del_location(project, pagenumber, pchange, location):
    "Deletes the item at the given location in the page, returns new pchange"
    return del_location_in_page(project, pagenumber, pchange, location)


def move_location(project, pagenumber, pchange, from_location, to_location):
    """Move an item in the given page from one spot to another, defined by its location
       Returns new page change uuid"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    if to_location == from_location:
        # no movement
        return pchange

    from_location_string, from_container, from_location_integers = from_location
    to_location_string, to_container, to_location_integers = to_location

    if (from_location_string != to_location_string) or (from_container != to_container):
        raise ServerError(message="Unable to move")

    location_string = from_location_string

    if from_container is None:
        # item to move is not in a widget
        if (location_string != 'head') and (location_string != 'body') and (location_string != 'svg'):
            raise ServerError(message="Unable to move item")
        # location integers are integers within the page, move it 
        up = False
        i = 0
        while True:
            if to_location_integers[i] < from_location_integers[i]:
                up = True
                break
            if to_location_integers[i] > from_location_integers[i]:
                # up is False
                break
            # so this digit is the same
            i += 1
            if len(from_location_integers) == i:
                # from location integers exhausted and equal to to_location! Trying to move an item under itself
                return pchange
            if len(to_location_integers) == i:
                up = True
                break
        if up:
            # move in the upwards direction
            #    delete it from original location
            #    insert it into new location
            try:
                # get top part such as head, body, svg
                top_part = page.location_item((location_string, None, ()))
                # get the part at from_location_integers
                part = top_part.get_location_value(from_location_integers)
                # delete part from top part location
                top_part.del_location_value(from_location_integers)
                # and insert it in the new location
                top_part.insert_location_value(to_location_integers, part)
            except Exception:
                raise ServerError(message="Unable to move item")
        else:
            # move in the downwards direction
            #    insert it into new location
            #    delete it from original location
            try:
                # get top part such as head, body, svg
                top_part = page.location_item((location_string, None, ()))
                # get the part at from_location_integers
                part = top_part.get_location_value(from_location_integers)
                # and insert it in the new location
                top_part.insert_location_value(to_location_integers, part)
                # delete part from current location
                top_part.del_location_value(from_location_integers)
            except Exception:
                raise ServerError(message="Unable to move item")
        # And save this page copy to the project
        return proj.save_page(page)

    # so item is in a widget, location_string is the widget name
    widget = page.widgets[location_string]
    ident_string = widget.ident_string

    # ident_string is of the form; project_pageidentnumber_body-x-y

    splitstring = ident_string.split("_")
    splitloc = splitstring[2].split("-")
    loc_top = splitloc[0]
    widg_ints = [ int(i) for i in splitloc[1:] ]

    widg_container_ints = list(widget.get_container_loc(from_container))

    from_location_ints = widg_ints + widg_container_ints + list(from_location_integers)
    to_location_ints = widg_ints + widg_container_ints + list(to_location_integers)

    # and move this item by calling this function again
    return move_location(project, pagenumber, pchange, (loc_top, None, from_location_ints), (loc_top, None, to_location_ints))


def css_style(project, pagenumber):
    "Return CSS style of a page, as an ordered Dictionary"
    proj, page = get_proj_page(project, pagenumber)
    if page.page_type != "CSS":
        raise ServerError(message = "Invalid page type")
    return page.style


def set_css_style(project, pagenumber, pchange, style):
    "Sets style in the CSS Page"
    # get a copy of the page, which can have a new style set
    # and can then be saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "CSS":
        raise ServerError(message = "Invalid page type")
    # Set the page style
    page.style = style
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def css_selector_properties(project, pagenumber, selector):
    "Return selctor properties of a CSS style of a page, as a string"
    proj, page = get_proj_page(project, pagenumber)
    if page.page_type != "CSS":
        raise ServerError(message = "Invalid page type")
    return page.selector_properties(selector)


def file_parameters(project, pagenumber):
    "Return filepath,mimetype of a page"
    proj, page = get_proj_page(project, pagenumber)
    if page.page_type != "FilePage":
        raise ServerError(message = "Invalid page type")
    return (page.filepath, page.mimetype)


def json_contents(project, pagenumber):
    "Return contents dictionary of a JSON page, where keys are WidgFieldInfo tuples"
    proj, page = get_proj_page(project, pagenumber)
    if page.page_type != "JSON":
        raise ServerError(message = "Invalid page type")
    contents = {}
    if page.content:
        for widgfield,value in page.content.items():
            wf = skiboot.make_widgfield(widgfield)
            if not wf:
                continue
            contents[WidgFieldInfo(wf.to_tuple_no_i(), str(wf), wf.to_str_tuple())] = value
    return contents


def remove_json_widgfield(project, pagenumber, pchange, str_widgfield):
    "Remove the given widgfield from the JSON page"
    # get a copy of the page, which can be edited
    # and then saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "JSON":
        raise ServerError(message = "Invalid page type")
    if str_widgfield in page.content:
        page.del_widgfield(str_widgfield)
    else:
        raise ServerError(message="Widgfield not recognised")
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def add_json_widgfield(project, pagenumber, pchange, str_widgfield, value):
    "Add the given widgfield and value to a JSON page"
    # get a copy of the page, which can be edited
    # and then saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "JSON":
        raise ServerError(message = "Invalid page type")
    page.add_widgfield(str_widgfield, value)
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def create_html_comment_in_page(project, pagenumber, pchange, location, text="comment here"):
    "Creates a new html comment in the given page, returns the new pchange and comment location"
    com = tag.Comment(text)
    # call skilift.insert_item_in_page to insert the item, save the page and return pchange
    new_pchange, new_location = insert_item_in_page(project, pagenumber, pchange, location, com)
    return new_pchange, new_location


def create_html_symbol_in_page(project, pagenumber, pchange, location, text="&nbsp;"):
    "Creates a new html symbol in the given page, returns the new pchange and symbol location"
    sym = tag.HTMLSymbol(text)
    # call skilift.insert_item_in_page to insert the item, save the page and return pchange
    new_pchange, new_location = insert_item_in_page(project, pagenumber, pchange, location, sym)
    return new_pchange, new_location


def create_html_element_in_page(project, pagenumber, pchange, location, name, brief, opentag = True):
    "Creates a new html element in the given page, returns the new pchange"
    if opentag:
        newpart = tag.Part(tag_name=name, brief=brief)
    else:
        newpart = tag.ClosedPart(tag_name=name, brief=brief)
    # call skilift.insert_item_in_page to insert the item, save the page and return pchange
    new_pchange, new_location = insert_item_in_page(project, pagenumber, pchange, location, newpart)
    return new_pchange


def create_part_in_page(project, pagenumber, pchange, location, json_data):
    """Builds the part from the given json string or ordered dictionary, and adds it to project either inserted into the html element
       currently at the given part location, or if not an element that can accept contents, inserted after the element.
       Returns new pchange value"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    if (page.page_type != "TemplatePage") and (page.page_type != "SVG"):
        raise ServerError(message = "Invalid page type")
    try:
        newpart = read_json.make_part_for_page(page, json_data)
    except Exception:
        raise ServerError("Unable to create part")
    # call skilift.insert_item_in_page to insert the item, save the page and return pchange
    new_pchange, new_location = insert_item_in_page(project, pagenumber, pchange, location, newpart)
    return new_pchange


def create_item_in_page(project, pagenumber, pchange, location, json_string):
    """Builds the item from the given json string, and adds it to project either inserted into the html element
       currently at the given location, or if not an element that can accept contents, inserted after the element.
       Returns new pchange value"""

    proj, page = get_proj_page(project, pagenumber, pchange)
    if (page.page_type != "TemplatePage") and (page.page_type != "SVG"):
        raise ServerError(message = "Invalid page type")
    try:
        newitem = read_json.make_item_for_page(page, json_string)
    except Exception:
        raise ServerError("Unable to create item")
    # call skilift.insert_item_in_page to insert the item, save the page and return pchange
    new_pchange, new_location = insert_item_in_page(project, pagenumber, pchange, location, newitem)
    return new_pchange


def page_element(project, pagenumber, pchange, location):
    """Return a PageElement tuple for the html element at the given location"""

    proj, page = get_proj_page(project, pagenumber, pchange)
    part = page.location_item(location)
    page_part = location[0]
    if location[1] is not None:
        # item is in a container in a widget, so location[0] will be the parent widget name
        widget = page.widgets[location[0]]
        if widget is not None:
           ident_top = widget.ident_string.split("-", 1)
           # ident_top[0] will be of the form proj_pagenum_head
           page_part = ident_top[0].split("_")[2]

    if part is None:
        raise ServerError("The item at the location has not been found")

    if hasattr(part, '__class__'):
        part_type = part.__class__.__name__
    else:
        raise ServerError("The item at the location must be an html element")

    if (part_type != "Part") and (part_type != "ClosedPart"):
        raise ServerError("The item at the location must be an html element")

    tag_name = part.tag_name
    brief = part.brief
    hide_if_empty = part.hide_if_empty
    attribs = OrderedDict(sorted(part.attribs.items(), key=lambda t: t[0]))

    return PageElement(project, pagenumber, pchange, location, page_part, part_type, tag_name, brief, hide_if_empty, attribs)


def edit_page_element(project, pagenumber, pchange, location, tag_name, brief, hide_if_empty, attribs):
    """Given an element at project, pagenumber, location
       sets the element values, returns page change uuid """
    proj, page = get_proj_page(project, pagenumber, pchange)
    part = page.location_item(location)
    part.tag_name = tag_name
    part.brief = brief
    part.attribs = OrderedDict(sorted(attribs.items(), key=lambda t: t[0]))
    if part.__class__.__name__ == "Part":
        part.hide_if_empty = hide_if_empty
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def del_attrib(project, pagenumber, pchange, location, attribute):
    """Given an element at project, pagenumber, location
       deletes the given attribute, returns page change uuid """
    proj, page = get_proj_page(project, pagenumber, pchange)
    part = page.location_item(location)
    part.del_one_attrib(attribute)
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def get_text(project, pagenumber, pchange, location):
    """Return a text string from the page at the given location"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    text = page.location_item(location)
    if not isinstance(text, str):
        raise ServerError("Item at this location is not identified as a text string")
    return text


def get_symbol(project, pagenumber, pchange, location):
    """Return a symbol from the page at the given location"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    sym = page.location_item(location)
    if not isinstance(sym, tag.HTMLSymbol):
        raise ServerError("Item at this location is not identified as a HTML Symbol")
    return sym.text


def edit_page_symbol(project, pagenumber, pchange, location, text):
    """Given a symbol at project, pagenumber, location
       sets the symbol, returns page change uuid """
    proj, page = get_proj_page(project, pagenumber, pchange)
    sym = page.location_item(location)
    if not isinstance(sym, tag.HTMLSymbol):
        raise ServerError("Item at this location is not identified as a HTML Symbol")
    sym.text = text
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def get_comment(project, pagenumber, pchange, location):
    """Return a comment from the page at the given location"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    com = page.location_item(location)
    if not isinstance(com, tag.Comment):
        raise ServerError("Item at this location is not identified as a HTML comment")
    return com.text


def edit_page_comment(project, pagenumber, pchange, location, text):
    """Given a comment at project, pagenumber, location
       sets the comment, returns page change uuid """
    proj, page = get_proj_page(project, pagenumber, pchange)
    com = page.location_item(location)
    if not isinstance(com, tag.Comment):
        raise ServerError("Item at this location is not identified as a HTML comment")
    com.text = text
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def page_textblock(project, pagenumber, pchange, location):
    """Return a PageTextBlock tuple for the textblock at the given location"""
    proj, page = get_proj_page(project, pagenumber, pchange)
    tblock = page.location_item(location)
    if not isinstance(tblock, tag.TextBlock):
        raise ServerError("Item at this location is not identified as a TextBlock")
    return PageTextBlock(project, pagenumber, pchange, location, tblock.textref, tblock.project, tblock.failmessage, tblock.escape, tblock.linebreaks)


def create_textblock_in_page(project, pagenumber, pchange, location, textref, failmessage, escape, linebreaks, tblock_project=''):
    "Creates a new textblock in the given page, returns the new pchange and location"
    tblock = tag.TextBlock(textref=textref, project=tblock_project, failmessage=failmessage, escape=escape, linebreaks=linebreaks)
    # call skilift.insert_item_in_page to insert the item, save the page and return pchange
    new_pchange, new_location = insert_item_in_page(project, pagenumber, pchange, location, tblock)
    return new_pchange, new_location


def edit_page_textblock(project, pagenumber, pchange, location, textref, tblock_project, failmessage, escape, linebreaks):
    """Given an textblock at project, pagenumber, location
       sets the textblock values, returns page change uuid """
    proj, page = get_proj_page(project, pagenumber, pchange)
    tblock = page.location_item(location)
    if not isinstance(tblock, tag.TextBlock):
        raise ServerError("Item at this location is not identified as a TextBlock")
    tblock.textref = textref
    tblock.failmessage = failmessage
    tblock.escape = escape
    tblock.linebreaks = linebreaks
    if project == tblock_project:
        tblock.project = ''
    else:
        tblock.project = tblock_project
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def page_filepath(project, pagenumber, pchange, filepath):
    "Sets filepath in the FilePage"
    # get a copy of the page, which can have a new filepath set
    # and can then be saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "FilePage":
        raise ServerError(message = "Invalid page type")
    # Set the page filepath
    page.filepath = filepath
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)


def page_mimetype(project, pagenumber, pchange, mimetype):
    "Sets mimetype in the FilePage"
    # get a copy of the page, which can have a new mimetype set
    # and can then be saved to the project
    proj, page = get_proj_page(project, pagenumber, pchange)
    if page.page_type != "FilePage":
        raise ServerError(message = "Invalid page type")
    # Set the page mimetype
    page.mimetype = mimetype
    # save the altered page, and return the page.change uuid
    return proj.save_page(page)




import setuptools

from distutils.core import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="skipole",
    version="3.5.9",
    author="Bernard Czenkusz",
    author_email="bernie@skipole.co.uk",
    description="A WSGI Application generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bernie-skipole.github.io/skipole/",
    packages=['skipole', 'skipole.skilift', 'skipole.ski', 'skipole.ski.responders', 'skipole.ski.validators', 'skipole.ski.widgets'],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application"
    ],
)





from .. import tag
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict



class SVGContainer(Widget):

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'width':FieldArg("text", "100", jsonset=True),
                        'height':FieldArg("text", "100", jsonset=True),
                        'viewBox':FieldArg("text", "", jsonset=True),
                        'preserveAspectRatio':FieldArg("text", "", jsonset=True)
                       }

    _container = ((),)

    def __init__(self, name=None, brief='', **field_args):
        """Acts as an SVG widget, containing other widgets, so show, class and dimensions can be set"""
        Widget.__init__(self, name=name, tag_name="svg", brief=brief, **field_args)
        self.htmlescaped = False
        self.linebreaks = False
        # the widget is a container
        self[0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("width"):
            self.update_attribs({"width":self.get_field_value("width")})
        if self.get_field_value("height"):
            self.update_attribs({"height":self.get_field_value("height")})
        if self.get_field_value("viewBox"):
            self.update_attribs({"viewBox":self.get_field_value("viewBox")})
        if self.get_field_value("preserveAspectRatio"):
            self.update_attribs({"preserveAspectRatio":self.get_field_value("preserveAspectRatio")})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<svg>  <!-- with widget id and class widget_class, and the given attributes -->
  <!-- further svg elements and widgets can be contained here -->
</svg>"""


class Group(Widget):

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'transform':FieldArg("text", "", jsonset=True)
                       }

    _container = ((),)

    def __init__(self, name=None, brief='', **field_args):
        """Acts as an g widget, containing other widgets, so group class, style, transform can be set"""
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self.htmlescaped = False
        self.linebreaks = False
        # The widget is a container
        self[0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
  <!-- further svg elements and widgets can be contained here -->
</g>"""


class TextBlockGroup(Widget):
    """A g tag, containing a TextBlock. The TextBlock is not escaped, so may contain
       svg commands which are set directly into the g"""

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {'textblock_ref':FieldArg("textblock_ref", ""),
                        'textblock_project':FieldArg("text", ""),
                        'content_refnotfound':FieldArg("text", ""),
                        'content_replaceblock':FieldArg("text", "" ,jsonset=True),
                        'transform':FieldArg("text", "", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        textblock_ref: The reference of the TextBlock appearing in the g
        textblock_project: Set with a project name if the TextBlock is defined in a sub project
        content_refnotfound: content to appear if the textblock is not found
        content_replaceblock: content set here will replace the textblock
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self[0] = ''
        self.htmlescaped = False
        self.linebreaks = False

    def _build(self, page, ident_list, environ, call_data, lang):
        # define the textblock
        tblock = self.get_field_value("textblock_ref")
        tblock.project = self.get_field_value('textblock_project')
        tblock.failmessage = self.get_field_value('content_refnotfound')
        tblock.escape = False
        tblock.linebreaks = False
        # place it at location 0
        if self.get_field_value("content_replaceblock"):
            self[0] = self.get_field_value("content_replaceblock")
        else:
            self[0] = tblock
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
   <!-- set with either content_replaceblock or textblock content as unescaped xtml -->
</g>"""



class TextGroup(Widget):
    """A g tag, containing a string. The string is not escaped, so may contain
       svg commands which are set directly into the g"""

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {
                        'text':FieldArg("text", "" ,jsonset=True),
                        'transform':FieldArg("text", "", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
       text: the string set here will be set as content
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="g", brief=brief, **field_args)
        self[0] = ''
        self.htmlescaped = False
        self.linebreaks = False

    def _build(self, page, ident_list, environ, call_data, lang):
        if self.get_field_value("text"):
            self[0] = self.get_field_value("text")
        if self.get_field_value("transform"):
            self.update_attribs({"transform":self.get_field_value("transform")})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<g>  <!-- with widget id and class widget_class, and transform attribute if given -->
   <!-- set with text content as unescaped xtml -->
</g>"""



class Rect(ClosedWidget):
    """An svg rect tag
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'x':FieldArg("text", "0", jsonset=True),
                        'y':FieldArg("text", "0", jsonset=True),
                        'rx':FieldArg("text", "0", jsonset=True),
                        'ry':FieldArg("text", "0", jsonset=True),
                        'width':FieldArg("text", "100", jsonset=True),
                        'height':FieldArg("text", "100", jsonset=True),
                        'fill':FieldArg("text", "none", jsonset=True),
                        'stroke':FieldArg("text", "black", jsonset=True),
                        'stroke_width':FieldArg("text", "1", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        x: x coords, for example 0
        y: y coords, for example 0
        rx: x radius, for example 0
        ry: y radius, for example 0
        width: A width field, for example 100
        height: A height field, for example 100
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge width
        """
        # pass fields to Widget
        ClosedWidget.__init__(self, name=name, tag_name="rect", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Set the attributes"
        # Note - all arg_descriptions are attributes, apart from stroke_width which is actually
        # attribute stroke-width. So update the tag with each item from arg_descriptions
        for att in self.arg_descriptions.keys():
            if self.get_field_value(att):
                if att == "stroke_width":
                    self.update_attribs({"stroke-width":self.get_field_value(att)})
                else:
                    self.update_attribs({att:self.get_field_value(att)})   

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<rect /> <!-- creates rectangle with widget id, class widget_class and the given attributes -->"""



class SimpleText(Widget):
    """An svg text tag
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'x':FieldArg("text", "0", jsonset=True),
                        'y':FieldArg("text", "0", jsonset=True),
                        'dx':FieldArg("text", "0", jsonset=True),
                        'dy':FieldArg("text", "0", jsonset=True),
                        'font_family':FieldArg("text", "Arial", jsonset=True),
                        'font_size':FieldArg("text", "20", jsonset=True),
                        'fill':FieldArg("text", "white", jsonset=True),
                        'stroke':FieldArg("text", "", jsonset=True),
                        'stroke_width':FieldArg("text", "", jsonset=True),
                        'text':FieldArg("text", "", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        x: x coords, for example 0
        y: y coords, for example 0
        dx: x displacement from absolute value, for example 0
        dy: y displacement from absolute value, for example 0
        font_family: font-family css  attribute, such as Arial
        font_size: font-size css  attribute, such as 20
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge width
        text: The text to display
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="text", brief=brief, **field_args)
        self.hide_if_empty=False


    def _build(self, page, ident_list, environ, call_data, lang):
        "Set the attributes"
        if self.get_field_value("x"):
            self.update_attribs({"x":self.get_field_value("x")})
        if self.get_field_value("y"):
            self.update_attribs({"y":self.get_field_value("y")})
        if self.get_field_value("dx"):
            self.update_attribs({"dx":self.get_field_value("dx")})
        if self.get_field_value("dy"):
            self.update_attribs({"dy":self.get_field_value("dy")})
        if self.get_field_value("font_family"):
            self.update_attribs({"font-family":self.get_field_value("font_family")})
        if self.get_field_value("font_size"):
            self.update_attribs({"font-size":self.get_field_value("font_size")})
        if self.get_field_value("fill"):
            self.update_attribs({"fill":self.get_field_value("fill")})
        if self.get_field_value("stroke"):
            self.update_attribs({"stroke":self.get_field_value("stroke")})
        if self.get_field_value("stroke_width"):
            self.update_attribs({"stroke-width":self.get_field_value("stroke_width")})
        self[0] = self.get_field_value("text")


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<text> <!-- with widget id, class widget_class and the given attributes -->
  <!-- contains given text -->
</text>"""



class TextList(Widget):
    """An svg text tag with further tspan elements
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'font_family':FieldArg("text", "Arial", jsonset=True),
                        'font_size':FieldArg("text", "20", jsonset=True),
                        'fill':FieldArg("text", "white", jsonset=True),
                        'stroke':FieldArg("text", "", jsonset=True),
                        'stroke_width':FieldArg("text", "", jsonset=True),
                        'lines': FieldArgTable(['integer', 'integer', 'text'])
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        font_family: font-family css  attribute, such as Arial
        font_size: font-size css  attribute, such as 20
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge width
        lines: List of lists, of [x,y,text]
        """
        # pass fields to Widget
        Widget.__init__(self, name=name, tag_name="text", brief=brief, **field_args)
        self.hide_if_empty=False


    def _build(self, page, ident_list, environ, call_data, lang):
        "Set the attributes"
        if self.get_field_value("font_family"):
            self.update_attribs({"font-family":self.get_field_value("font_family")})
        if self.get_field_value("font_size"):
            self.update_attribs({"font-size":self.get_field_value("font_size")})
        if self.get_field_value("fill"):
            self.update_attribs({"fill":self.get_field_value("fill")})
        if self.get_field_value("stroke"):
            self.update_attribs({"stroke":self.get_field_value("stroke")})
        if self.get_field_value("stroke_width"):
            self.update_attribs({"stroke-width":self.get_field_value("stroke_width")})
        lines = self.get_field_value("lines")
        if not lines:
            return
        line1 = lines[0]
        self.update_attribs({"x":str(line1[0])})
        self.update_attribs({"y":str(line1[1])})
        self[0] = line1[2]
        for line in lines[1:]:
            tspan = tag.Part(tag_name='tspan', text = line[2])
            tspan.update_attribs({"x":str(line[0])})
            tspan.update_attribs({"y":str(line[1])})
            self.append(tspan)

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<text> <!-- with widget id, class widget_class and the given attributes -->
  <!-- contains given text, each line within tspan elements -->
</text>"""




class Circle(ClosedWidget):
    """An svg circle tag
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'cx':FieldArg("text", "50", jsonset=True),
                        'cy':FieldArg("text", "50", jsonset=True),
                        'r':FieldArg("text", "40", jsonset=True),
                        'fill':FieldArg("text", "none", jsonset=True),
                        'stroke':FieldArg("text", "black", jsonset=True),
                        'stroke_width':FieldArg("text", "1", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        cx: centre x coords, for example 0
        cy: centre y coords, for example 0
        r: radius, for example 0
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge width
        """
        # pass fields to Widget
        ClosedWidget.__init__(self, name=name, tag_name="circle", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Set the attributes"
        # Note - all arg_descriptions are attributes, apart from stroke_width which is actually
        # attribute stroke-width. So update the tag with each item from arg_descriptions
        for att in self.arg_descriptions.keys():
            if self.get_field_value(att):
                if att == "stroke_width":
                    self.update_attribs({"stroke-width":self.get_field_value(att)})
                else:
                    self.update_attribs({att:self.get_field_value(att)})   

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<circle /> <!-- creates circle with widget id, class widget_class and the given attributes -->"""


class Line(ClosedWidget):
    """An svg line tag
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'x1':FieldArg("text", "0", jsonset=True),
                        'y1':FieldArg("text", "0", jsonset=True),
                        'x2':FieldArg("text", "100", jsonset=True),
                        'y2':FieldArg("text", "100", jsonset=True),
                        'stroke':FieldArg("text", "black", jsonset=True),
                        'stroke_width':FieldArg("text", "1", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        x1,y1: coordinates of line start
        x2,y2: coordinates of line end
        stroke: The line colour
        stroke_width: The outline edge width
        """
        # pass fields to Widget
        ClosedWidget.__init__(self, name=name, tag_name="line", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Set the attributes"
        # Note - all arg_descriptions are attributes, apart from stroke_width which is actually
        # attribute stroke-width. So update the tag with each item from arg_descriptions
        for att in self.arg_descriptions.keys():
            if self.get_field_value(att):
                if att == "stroke_width":
                    self.update_attribs({"stroke-width":self.get_field_value(att)})
                else:
                    self.update_attribs({att:self.get_field_value(att)})   

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<line /> <!-- creates line with widget id, class widget_class and the given attributes -->"""


class Ellipse(ClosedWidget):
    """An svg ellipse tag
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'cx':FieldArg("text", "50", jsonset=True),
                        'cy':FieldArg("text", "50", jsonset=True),
                        'rx':FieldArg("text", "40", jsonset=True),
                        'ry':FieldArg("text", "20", jsonset=True),
                        'fill':FieldArg("text", "none", jsonset=True),
                        'stroke':FieldArg("text", "black", jsonset=True),
                        'stroke_width':FieldArg("text", "1", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        cx: centre x coords, for example 50
        cy: centre y coords, for example 50
        rx: the horizontal radius, for example 40
        ry: the vertical radius, for example 20
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge width
        """
        # pass fields to Widget
        ClosedWidget.__init__(self, name=name, tag_name="ellipse", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Set the attributes"
        # Note - all arg_descriptions are attributes, apart from stroke_width which is actually
        # attribute stroke-width. So update the tag with each item from arg_descriptions
        for att in self.arg_descriptions.keys():
            if self.get_field_value(att):
                if att == "stroke_width":
                    self.update_attribs({"stroke-width":self.get_field_value(att)})
                else:
                    self.update_attribs({att:self.get_field_value(att)})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<ellipse /> <!-- creates ellipse with widget id, class widget_class and the given attributes -->"""



class Polygon(ClosedWidget):
    """An svg polygon tag
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'points':FieldArgTable(['text', 'text'], jsonset=True),
                        'fill':FieldArg("text", "none", jsonset=True),
                        'stroke':FieldArg("text", "black", jsonset=True),
                        'stroke_width':FieldArg("text", "1", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        points : a list of two element lists, each inner list being a point
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge width
        """
        # pass fields to Widget
        ClosedWidget.__init__(self, name=name, tag_name="polygon", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Set the attributes"

        if self.get_field_value("fill"):
            self.update_attribs({"fill":self.get_field_value("fill")})
        if self.get_field_value("stroke"):
            self.update_attribs({"stroke":self.get_field_value("stroke")})
        if self.get_field_value("stroke_width"):
            self.update_attribs({"stroke-width":self.get_field_value("stroke_width")})

        if not self.get_field_value('points'):
            return

        points_att = ""
        points = self.get_field_value('points')
        for x,y in points:
            points_att += "%s,%s " % (x,y)
        self.update_attribs({'points':points_att})   
  

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<polygon /> <!-- creates polygon with widget id, class widget_class and the given points -->"""



class Polyline(ClosedWidget):
    """An svg polyline tag
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {
                        'points':FieldArgTable(['text', 'text'], jsonset=True),
                        'fill':FieldArg("text", "none", jsonset=True),
                        'stroke':FieldArg("text", "black", jsonset=True),
                        'stroke_width':FieldArg("text", "1", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        points : a list of two element lists, each inner list being a point
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge width
        """
        # pass fields to Widget
        ClosedWidget.__init__(self, name=name, tag_name="polyline", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Set the attributes"

        if self.get_field_value("fill"):
            self.update_attribs({"fill":self.get_field_value("fill")})
        if self.get_field_value("stroke"):
            self.update_attribs({"stroke":self.get_field_value("stroke")})
        if self.get_field_value("stroke_width"):
            self.update_attribs({"stroke-width":self.get_field_value("stroke_width")})

        if not self.get_field_value('points'):
            return

        points_att = ""
        points = self.get_field_value('points')
        for x,y in points:
            points_att += "%s,%s " % (x,y)
        self.update_attribs({'points':points_att})   
  

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<polyline /> <!-- creates polyline with widget id, class widget_class and the given points -->"""



class Path(ClosedWidget):
    """An svg path tag
    """

    # This class does not display any error messages
    display_errors = False

    arg_descriptions = {'d':FieldArg("text", "", jsonset=True),
                        'fill':FieldArg("text", "none", jsonset=True),
                        'stroke':FieldArg("text", "black", jsonset=True),
                        'stroke_width':FieldArg("text", "1", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        d: The path data
        fill: The fill colour, use none for no fill
        stroke: The outline edge colour
        stroke_width: The outline edge width
        """
        # pass fields to Widget
        ClosedWidget.__init__(self, name=name, tag_name="path", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Set the attributes"
        # Note - all arg_descriptions are attributes, apart from stroke_width which is actually
        # attribute stroke-width. So update the tag with each item from arg_descriptions
        for att in self.arg_descriptions.keys():
            if self.get_field_value(att):
                if att == "stroke_width":
                    self.update_attribs({"stroke-width":self.get_field_value(att)})
                else:
                    self.update_attribs({att:self.get_field_value(att)})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<path /> <!-- creates path with widget id, class widget_class and the given attributes -->"""


